'''
 Copyright (c) 2019, UChicago Argonne, LLC
 See LICENSE file.
'''
import PyQt5.QtCore as qtCore
from PyQt5.QtWidgets import QVBoxLayout, QSlider, QLabel, QSplitter, \
    QWidget, QMenu, QAction
from PyQt5.QtCore import pyqtSlot, pyqtSignal, QRectF
from minixs import METHOD_ENTER_STR
import numpy as np
from PIL import Image
from pyqtgraph import ImageItem, ScatterPlotItem, functions
from pyqtgraph import GraphicsLayoutWidget
import os
import logging
from minixs.gui.qthelpers.qtfilterview import REGISTRY
from pyqtgraph.widgets.HistogramLUTWidget import HistogramLUTWidget
from PyQt5.Qt import QRadioButton, QGroupBox, QPushButton, QCheckBox, QComboBox,\
    QHBoxLayout, QMessageBox
from minixs.constants import DIRECTION_NAMES
from minixs.filter import MIN_VISIBLE_FILTER, MAX_VISIBLE_FILTER,\
    get_filter_by_name
from pyqtgraph.graphicsItems.ROI import RectROI
from matplotlib import cm
from minixs.gui.qthelpers import qtfilterview

logger = logging.getLogger(__name__)

FIND_CRYSTAL_SIGNAL = "findCrystals"
SHOW_CRYSTAL_SIGNAL = "showCrystals"
VIEW_CHOICE_CHANGED_SIGNAL = "viewChoiceChanged"
NO_EXPOSURES_LABEL = "No Exposures Loaded"
CALIBRATION_MATRIX_LABEL = "CalibrationMatrix"
EXPOSURE_LABEL_TEMPLATE = "%d/%d %s %f"
CHECK_BOX_KEY = 'checkBox'
VALUE_KEY = 'value'

class ImageViewQt(QWidget):
    '''
    classdocs
    '''
    findCrystals = pyqtSignal(name=FIND_CRYSTAL_SIGNAL)
#     showCrystals = pyqtSignal(bool, name=SHOW_CRYSTAL_SIGNAL)

    def __init__(self, *args, **kwargs):
        '''
        Constructor
        '''
        super(ImageViewQt, self).__init__(*args, **kwargs)
        layout = QVBoxLayout()
        hSplitter = QSplitter()
        
        self.filters = FilterBox()
        self.imageView = ImageBox()
        self.matrixView = CalibrationMatrixBox()
        self.tools = ToolBox()
        self.filters.filterChanged[object].connect(self.handleImageFilterChanged)

        self.imageBox=QWidget()
        imageBoxLayout = QHBoxLayout()
        imageBoxLayout.addWidget(self.imageView)
        self.imageBox.setLayout(imageBoxLayout)
        
        hSplitter.addWidget(self.filters)
        hSplitter.addWidget(self.imageBox)
        
        hSplitter.addWidget(self.tools)
        
        self.tools.findCrystals.connect(self.handleFindCrystals)
        self.tools.calibrationRadioButton.toggled.connect(
            self.handleViewChangeToCalibration)
        self.tools.exposureRadioButton.toggled.connect(
            self.handleViewChangeToExposure)
        self.tools.showCrystals[bool].connect(self.imageView.showCrystals)
        
        layout.addWidget(hSplitter)
        self.calibrateButton = QPushButton("Calibrate")
        layout.addWidget(self.calibrateButton)
        self.setLayout(layout)
        
    def deleteImages(self, rows):
        logger.debug(METHOD_ENTER_STR, (rows,))
        self.imageView.deleteImages(rows)
        
    def loadImages(self, filenames, energies):
        logger.debug(METHOD_ENTER_STR % ((filenames,energies),))
        self.imageView.loadImages(filenames, energies)
        self.filters.setEnabled(True)
        
    def handleImageFilterChanged(self, filterObj):
        logger.debug(METHOD_ENTER_STR % filterObj)
        logger.debug("filter %s enabled: %s" % (filterObj.text(), filterObj.filterEnabled()))
        logger.debug("filter %s value: %s" % (filterObj.text(), filterObj.getValue()))
        if self.imageView.imageStack.size == 0:  # Cannot do anything with no data.
            return
        self.updateImageDisplayFilters()
        
    def updateImageDisplayFilters(self):
        imageDisplayFilters = self.filters.getImageDisplayFilters()
        self.imageView.applyFilters(self.filters.getFilters())
        self.imageView.applyImageFilters(imageDisplayFilters)
        
    def handleFindCrystals(self):
        self.findCrystals.emit()
        
    def handleViewChangeToExposure(self, checked):
        logger.debug(METHOD_ENTER_STR % checked)
        if checked:
            imageBoxLayout = self.imageBox.layout()
            self.matrixView.setVisible(False)
            imageBoxLayout.removeWidget(self.matrixView)
            self.imageView.setVisible(True)
            imageBoxLayout.addWidget(self.imageView)
            imageBoxLayout.update()

    def handleViewChangeToCalibration(self, checked):
        logger.debug(METHOD_ENTER_STR % checked)
        if checked:
            imageBoxLayout = self.imageBox.layout()
            self.imageView.setVisible(False)
            imageBoxLayout.removeWidget(self.imageView)
            self.matrixView.setVisible(True)
            imageBoxLayout.addWidget(self.matrixView)
            imageBoxLayout.update()
        
class FilterBox(QWidget):
    filterChanged = pyqtSignal(object, name="filterChanged")
    def __init__(self, *args, **kwargs):
        super(FilterBox, self).__init__(*args, **kwargs)
        self.views = {}
        filters = REGISTRY
        self.filterWidgets = []
        logger.debug("Filters: %s" % filters)
        layout = QVBoxLayout()
        for dfilter in filters:
            filterWidget = dfilter[1](filter=dfilter[0]())
            layout.addWidget(filterWidget)
            filterWidget.filterChanged[object].connect(self.filterStateChanged)
            self.filterWidgets.append(filterWidget)
            self.views[dfilter[0]] = filterWidget
        label = QLabel("Dispersive Direction")
        self.dispersiveDirection = QComboBox()
        self.dispersiveDirection.addItems(DIRECTION_NAMES)
        hLayout = QHBoxLayout()
        hLayout.addWidget(label)
        hLayout.addWidget(self.dispersiveDirection)
        layout.addLayout(hLayout)
        self.setLayout(layout)
        
        self.setEnabled(False)
    
    def filterStateChanged(self, filterObj):
        self.filterChanged.emit(filterObj)

    def _getDisplayFilterVals(self, filterWidget):
        return {CHECK_BOX_KEY: filterWidget.filterEnabled(),
                VALUE_KEY: filterWidget.getValue()}
                    
    def getFilters(self):
        filters = []
        for fw in self.filterWidgets:
            filters.append((fw.filter.name, fw.filterEnabled(), fw.getValue()))
        return filters
            
    def _getFilterDefaults(self, fltr):
        valStr = ""
        
        if hasattr(fltr, 'default_name'):
            valStr = fltr.val_to_str(fltr.default_val)
        val = fltr.str_to_val(valStr)
          
        enabled = False
        if hasattr(fltr, 'default_enabled'):
            enabled = fltr.default_enabled
        
        return (enabled, val)
            
        
        
    def getImageDisplayFilters(self):
        displayFilters = {}
        for dfilter in self.filterWidgets:
            logger.debug("DisplayFilter %s" % dfilter)
            if dfilter.filter.name == MIN_VISIBLE_FILTER:
                displayFilters[MIN_VISIBLE_FILTER] = self._getDisplayFilterVals(dfilter)
            elif  dfilter.filter.name == MAX_VISIBLE_FILTER:
                displayFilters[MAX_VISIBLE_FILTER] = self._getDisplayFilterVals(dfilter)
        return displayFilters
        
    def setFilters(self, filters):
        logger.debug(METHOD_ENTER_STR)
        fmap = {}
        for f in filters:
            fmap[f.__class__] = f
            
        for ftype, view_class in qtfilterview.REGISTRY:
            if ftype in fmap:
                enabled = True
                val = fmap[ftype].get_val()
            else:
                enabled, val = self._getFilterDefaults(ftype)
                enabled = False
            self.views[ftype].filterEnableChanged(enabled)
            if val is not None:
                self.views[ftype].setValue(val)
        
        
class ImageBox(QWidget):
    '''
    Widget box to display the image stack, & slider to allow switching 
    which image is displayed
    '''    
    allImagesDeleted = pyqtSignal(name="allImagesDeleted")
    
    def __init__(self, *args, **kwargs):
        super(ImageBox, self).__init__(*args, **kwargs)
        self.lastMousePosition = [0, 0]
        self.handlersConnected = False
        self.images = []
        self.imageEnergies = []
        self.imageStack = None
        self.rois = []
        self.filters = []
        
        layout = QVBoxLayout()
        self.imageLabel=QLabel(NO_EXPOSURES_LABEL)
        layout.addWidget(self.imageLabel)
        graphicsWidget = GraphicsLayoutWidget()
        self.viewBox = graphicsWidget.addViewBox()
        self.viewBox.setAspectLocked(True)
        self.updateMenu()
        self.imageView = ImageItem( border={'color':'333'})
        self.imageView.setOpts(axisOrder='row-major')
        self.viewBox.setVisible(False)
        self.imageNumber = -1
        self.fitPointsView = ScatterPlotItem()
        
        
        self.viewBox.addItem(self.imageView)
        self.viewBox.show()
        layout.addWidget(graphicsWidget)
        self.slider = QSlider(qtCore.Qt.Horizontal)
        layout.addWidget(self.slider)
        
        self.histWidget = HistogramLUTWidget(image=self.imageView)
        colormap = cm.get_cmap("Greys_r")
        colormap._init()
        lut = (colormap._lut *255).view(np.ndarray)
        self.imageView.setLookupTable(lut)
        self.histWidget.item.setImageItem(self.imageView)
        #layout.addWidget(self.histWidget)
     
        self.setLayout(layout)


    def _addDecoratedROI(self, position, size):
        imageSize = list(self.imageStack[0].shape)
        bounds = QRectF(0, 0, imageSize[1], imageSize[0])
        roi = RectROI(position, size,\
                      maxBounds=bounds,
                      removable=True, movable=True,
                      sideScalers=True)
        roi.sigRemoveRequested.connect(self.removeROI)
        roi.addScaleHandle([0,.5], [1,.5])
        roi.addScaleHandle([.5,0], [.5,1])
        self.rois.append(roi)
        self.viewBox.addItem(roi)
        
    def addROI(self):
        '''
        Add a single ROI at [0,0] & size[50, 50].  This is 
        meant to add an ROI that will be moved to a desired location
        by the user.
        '''
        if self.imageStack is not None:
            imageSize = list(self.imageStack[0].shape)
            logger.debug("imageSize %s" % (imageSize,))
            self._addDecoratedROI(self.lastMousePosition, [50, 50])
        else:
            self._addDecoratedROI(self.lastMousePosition, [50, 50])

    def addROIs(self, rects):
        '''
        Add ROIs of size & location determined by the list
        of rects passed in
        '''
        for rect in rects:
            logger.debug("rect %s" % rect)
            position = rect[0]
            size = [rect[1][0] - rect[0][0],\
                                    rect[1][1] - rect[0][1]]
            self._addDecoratedROI(position, size)
            
    
    def applyFilters(self, filters):
        '''
        Apply filters as defined on the current image.  These filters will 
        maks out data based on criteria specified in the filters.
        '''
        logger.debug(METHOD_ENTER_STR % (filters,))
        self.filters = filters
        filteredData = self.getFilteredData()
        imageLevels = self.imageView.getLevels()
        logger.debug("Display image %s" % (imageLevels,))
        logger.debug("filtered Data min/max %s/%s" % (filteredData.min(), filteredData.max()))
        if (filteredData.min() != filteredData.max()):
            self.imageView.setImage(filteredData, autoLevels=False)
            self.imageView.border = functions.mkPen({'color': '333'})
            self.viewBox.update()
        else:         # pyqt graph does not crashes if image is completely flat
            self.imageView.setImage(None)
            self.imageView.border = functions.mkPen({'color': 'F00'})
            self.viewBox.update()

    def applyImageFilters(self, filters):
        '''
        Use the filter values for Min Visible and
        Max Visible to scale the image by setting the level on the 
        current image.
        '''
        logger.debug(METHOD_ENTER_STR % (filters,))
        if MIN_VISIBLE_FILTER not in filters.keys():
            logger.warning("Minimum Visible Filter is missing")
            return
        if MAX_VISIBLE_FILTER not in filters.keys():
            logger.warning("Maximum Visible Filter is missing")
            return
        minRange = 0
        maxRange = 1000
        imageNumber = self.slider.value()
        logger.debug("Filters %s" % filters)
        if filters[MIN_VISIBLE_FILTER][CHECK_BOX_KEY]:
            minRange = filters[MIN_VISIBLE_FILTER][VALUE_KEY]
        else:
            minRange = np.min(self.imageStack[imageNumber,:,:])
        if filters[MAX_VISIBLE_FILTER][CHECK_BOX_KEY]:
            maxRange = filters[MAX_VISIBLE_FILTER][VALUE_KEY]
        else:
            maxRange = np.min(self.imageStack[imageNumber,:,:])
        self.histWidget.item.setLevels(minRange, maxRange)
        self.imageView.setLevels((minRange,maxRange))

    def clearImages(self):
        '''
        if the images or energies are cleared then clear the imageView image
        and change the imageLabel.  Alsp  clear out the image names, imageStack
        and Energu List.
        '''
        self.imageLabel(NO_EXPOSURES_LABEL)
        self.imageView.setImage(None)
        
    def connectDataSignals(self):
        if not self.handlersConnected and len(self.images) >0:
            self.histWidget.item.sigLevelChangeFinished.connect( \
                                    self.handleSigLevelChangeFinished)
            self.histWidget.item.sigLookupTableChanged.connect( \
                                    self.handleSigLookupTableChanged)
#                 self.histWidget.item.sigLevelsChanged.connect( \
#                                         self.handleSigLevelsChanged)
            self.imageView.scene().sigMouseMoved.connect( \
                                    self.handleMouseMove)
            self.slider.valueChanged.connect(self.setCurrentImage)
            self.handlersConnected = True
            
        
        
    def deleteAllImages(self):
        imageRows = range(len(self.images))
        self.deleteImages(reversed(imageRows))
        self.allImagesDeleted.emit()
        
    def deleteImages(self, rows):
        '''
        delete a set of images from the display,  This is triggerd by the 
        delete rows from the exposure selector.
        '''
        logger.debug(METHOD_ENTER_STR % (rows,))
        self.disconnectDataSignals()
        for row in rows:
            logger.debug("Removing Row %s" % row)
            if len(self.images) > 0:
                self.images.remove(self.images[row])
                self.imageStack = np.delete(self.imageStack, row, axis=0 )
            if len(self.imageEnergies) > 0:
                self.imageEnergies.remove(self.imageEnergies[row])
            if len(self.images) > 0:
                self.setCurrentImage(0)
                logger.debug("Setting maximum value for the slider to %s" %
                         (self.imageStack.shape[0] - 1))
                self.slider.setMaximum(self.imageStack.shape[0] - 1)
                self.connectDataSignals()
            else:
                self.setCurrentImage(None)
                self.slider.setMaximum(self.slider.minimum())
                self.disconnectDataSignals()
                
    def disconnectDataSignals(self):
        '''
        Disconnect signals related to selecting or changing data, ranges, etc
        while data is being loaded, deleted, modified etc to avoid problems.
        '''
        if self.handlersConnected:
            self.histWidget.item.sigLevelChangeFinished.disconnect( \
                                        self.handleSigLevelChangeFinished)
            self.histWidget.item.sigLookupTableChanged.disconnect( \
                                        self.handleSigLookupTableChanged)
#             self.histWidget.item.sigLevelsChanged.disconnect( \
#                                         self.handleSigLevelsChanged)
            self.handlersConnected = False
            self.imageView.scene().sigMouseMoved.disconnect( \
                                        self.handleMouseMove)
            self.slider.valueChanged.disconnect(self.setCurrentImage)
        
        
    def getFilteredData(self):
        logger.debug(METHOD_ENTER_STR)
        filteredData = np.copy(self.imageStack[self.imageNumber, slice(None), slice(None)]).astype(np.float64)
        for (name, enabled, val ) in self.filters:
            if enabled:
                try:
                    fltr = get_filter_by_name(name)
                    fltr.set_val(val)
                    fltr.filter(filteredData, 0.0)
                except ValueError:
                    logger.debug("trouble with value change in fillter: %s" \
                                 % fltr)
        return filteredData
        
                       
    def handleMouseMove(self, viewPos):
        '''
        As the mouse is moved over the image show the image possition in the 
        status bar
        '''
#         logger.debug("mouseMoved in %s", object)
        imageNumber = self.slider.value()
        mainWindow = self.window()
        scenePos = self.imageView.mapFromScene(viewPos)
#         logger.debug ("scenePos %s" % ((scenePos,)))
        row, col = int(scenePos.y()), int(scenePos.x())
#         logger.debug ("row, col %s,%s" % (row, col))
        image = self.getFilteredData()
#         logger.debug ("image.shape %s" % ((image.shape,)))
        if (0 <= row < image.shape[0]) and (0 <= col < image.shape[1]):
            mainWindow.statusBar().showMessage("x=%s, y=%s, val=%s" % (col, 
                                                       row, 
                                                       image[row,col]))
            self.lastMousePosition = [col, row]
        else:
            mainWindow.statusBar().showMessage("")

    def handleSigLookupTableChanged(self, obj):
        '''
        '''
        logger.debug(METHOD_ENTER_STR % obj)
        imageNumber = self.slider.value()
        lut = self.histWidget.item.getLookupTable( \
                                    img=self.imageStack[imageNumber,:,:])
        logger.debug("new Lookup Table: %s" % ((lut,)))
          
    def handleSigLevelsChanged(self, obj):
        '''
        Handler for changing the signal levels that define
        the range of data for the image
        '''
        logger.debug(METHOD_ENTER_STR % obj)
        levels = self.histWidget.getLevels()
        logger.debug("new Levels min,max %s, %s" % levels)
  
    def handleSigLevelChangeFinished(self, obj):
        '''
        Handler for when finished changing the signal
        levels that define the limits for the colormap
        used to display the data
        '''
        logger.debug(METHOD_ENTER_STR % obj)
        levels = self.histWidget.getLevels()
        logger.debug("new Levels min,max %s, %s" % levels)

    def loadImages(self, filenames, energies):
        '''
        After images are selected load images into a stack and display the
        first image.
        '''
        logger.debug(METHOD_ENTER_STR, (filenames,))
        self.disconnectDataSignals()
        self.images = filenames
        self.imageEnergies = energies
        #self.imgArray = np.empty()
        self.imageStack = np.array([])
        try:
            if len(filenames) == 0:
                return
            imgs = [Image.open(image, "r") for image in filenames]
            if len(imgs) > 0: 
                logger.debug("images %s" % imgs)
                self.imageStack = np.stack(imgs, axis=0)
#             self.imageStack = np.swapaxes(self.imageStack, 1, 2)
                logger.debug("stack shape %s" % (self.imageStack.shape,))
                self.slider.setMinimum(0)
                self.slider.setMaximum(self.imageStack.shape[0]-1)
                self.setCurrentImage(0)
            else:
                self.slider.setMinimum(0)
                self.slider.setMaximum(self.imageStack.shape[0]-1)
                self.setCurrentImage(None)
                return
            self.connectDataSignals()
            self.viewBox.autoRange(items=[self.imageView,])
            self.imageView.update()
                    
        except IOError as ex:
            msg = str(ex)
            msg = msg + "\nSelected file does not seem to be an image."
            QMessageBox.warning(self, self.tr("Calibrator"), self.tr(msg))
            
    def removeAllROIs(self):
        for ii in reversed(range(len(self.rois))):
            self.removeROI(self.rois[ii])
        
    def removeROI(self, roi):
        '''
        Remove the highlighted ROI
        '''
        roi.sigRemoveRequested.disconnect(self.removeROI)
        self.viewBox.removeItem(roi)
        self.rois.remove(roi)
            
    def replaceXtals(self, rects):
        self.removeAllROIs()
        self.addROIs(rects)
        
        
    @pyqtSlot(int)
    def setCurrentImage(self, imageNumber):
        '''
        As the selected image number changes, display the image and change the imageLabel
        '''
        logger.debug("self.images %s "  % self.images)
        logger.debug("self.imageEnergies %s "  % self.imageEnergies)
        logger.debug("self.imageStack %s "  % self.imageStack)
        
        if not imageNumber is None:
            self.imageLabel.setText(EXPOSURE_LABEL_TEMPLATE % (imageNumber,
                                                     self.imageStack.shape[0] -1,
                                                     os.path.basename(self.images[imageNumber]),
                                                     self.imageEnergies[imageNumber]))
            logger.debug("Image Levels min, max: %s", self.imageView.getLevels())
            colormap = cm.get_cmap("Greys_r")
            colormap._init()
            lut = (colormap._lut *255).view(np.ndarray)
            self.imageView.setLookupTable(lut)
            self.imageNumber = imageNumber
            filteredData = self.getFilteredData()
            self.imageView.setImage(filteredData, autoLevels=False)
            self.viewBox.setVisible(True)
        else:
            self.imageLabel.setText(NO_EXPOSURES_LABEL)
            self.imageView.setImage(image=None)
            self.viewBox.setVisible(False)
#         self.histWidget.item.setImageItem(self.imageStack[imageNumber,:,:])
#         self.histWidget.item.setImageItem(self.imageView)
            
    @pyqtSlot(bool)
    def showCrystals(self, show):
        '''
        Toggle display of ROIs for the crystal boundaries for the 
        image
        '''
        logger.debug(METHOD_ENTER_STR % show)
        for roi in self.rois:
            if show:
                roi.show()
            else:
                roi.hide()

    def setFilters(self,filters):
        self.filters = filters
        
        
    def updateMenu(self):
        '''
        '''
        menu = self.viewBox.menu
        self.roiMenu = QMenu("ROIs")
        
        addROI = QAction("Add ROI", self.roiMenu)
        self.roiMenu.addAction(addROI)
        addROI.triggered.connect(self.addROI)
        removeAllROIs = QAction("Remove All ROIs", self.roiMenu)
        self.roiMenu.addAction(removeAllROIs)
        removeAllROIs.triggered.connect(self.removeAllROIs)
        menu.addMenu(self.roiMenu)
    
class CalibrationMatrixBox(QWidget):
    
    def __init__(self, *args, **kwargs):
        super(CalibrationMatrixBox, self).__init__(*args, **kwargs)
        self.handlersConnected = False
        self.image = None
        self.calImage = None
        self.lastMousePosition = [0,0]
        layout = QVBoxLayout()
        self.imageLabel=QLabel(CALIBRATION_MATRIX_LABEL)
        layout.addWidget(self.imageLabel)
        graphicsWidget = GraphicsLayoutWidget()
        self.viewBox = graphicsWidget.addViewBox()
        self.viewBox.setAspectLocked(True)
        self.imageView = ImageItem( border={'color':'333'})
        self.imageView.setOpts(axisOrder="row-major")
        self.viewBox.addItem(self.imageView)
        self.viewBox.show()
        layout.addWidget(graphicsWidget)
        
        self.histWidget = HistogramLUTWidget(image=self.imageView)
        self.histWidget.item.setImageItem(self.imageView)
#         layout.addWidget(self.histWidget)
        
        self.setLayout(layout)
        self.setVisible(False)
        
    def setImage(self, image):
        '''
        Set the image to be displayed for the calibrated data.
        '''
        if self.handlersConnected:
            self.imageView.scene().sigMouseMoved.disconnect( \
                                        self.handleMouseMove)
            self.handlersConnected = False
        self.image = image
        self.calImage = np.array(image)
#         self.calImage = np.swapaxes(self.calImage, 0, 1)
        nonZeroVals = np.nonzero(self.calImage)
        minVal = np.min(self.calImage[nonZeroVals])
        maxVal = np.max(self.calImage[nonZeroVals])
        # colormap setup
        colormap = cm.get_cmap("jet")
        colormap._init()
        lut = (colormap._lut *255).view(np.ndarray)
        self.imageView.setLookupTable(lut)
        self.imageView.setImage(self.calImage, levels=[minVal,maxVal])
        if not self.handlersConnected:
            self.imageView.scene().sigMouseMoved.connect( \
                                        self.handleMouseMove)
            self.handlersConnected = True

    def handleMouseMove(self, viewPos):
        '''
        As the mouse is moved over the image show the image possition in the 
        status bar
        '''
#         logger.debug("mouseMoved in %s", object)
        mainWindow = self.window()
        scenePos = self.imageView.mapFromScene(viewPos)
#         logger.debug ("scenePos %s" % ((scenePos,)))
        row, col = int(scenePos.y()), int(scenePos.x())
#         logger.debug ("row, col %s,%s" % (row, col))
        image = self.calImage
#         logger.debug ("image.shape %s" % ((image.shape,)))
        if (0 <= row < image.shape[0]) and (0 <= col < image.shape[1]):
            mainWindow.statusBar().showMessage("x=%s, y=%s, val=%s" % (col, 
                                                       row, 
                                                       image[row,col]))
            self.lastMousePosition = [col, row]
        else:
            mainWindow.statusBar().showMessage("")
        
        
class ToolBox(QWidget):
    findCrystals = pyqtSignal(name=FIND_CRYSTAL_SIGNAL)
    showCrystals = pyqtSignal(bool, name=SHOW_CRYSTAL_SIGNAL)
    viewChoiceChanged = pyqtSignal(name=VIEW_CHOICE_CHANGED_SIGNAL)
    def __init__(self, *args, **kwargs):
        super(ToolBox, self).__init__(*args, **kwargs)
        self.rects = []
        layout = QVBoxLayout()
        choices = ["Exposures", "Calibration Matrix"]
        self.radioGroup = QGroupBox("View", maximumHeight=100)
        radioLayout = QVBoxLayout()
        self.exposureRadioButton = QRadioButton(choices[0])
        self.exposureRadioButton.setChecked(True)
        self.calibrationRadioButton = QRadioButton(choices[1])
        radioLayout.addWidget(self.exposureRadioButton, stretch=0)
        radioLayout.addWidget(self.calibrationRadioButton, stretch=0)
        
        self.radioGroup.setLayout(radioLayout)
        findCrystalsButton = QPushButton("Find Crystal Bounds", maximumHeight=50)
        findCrystalsButton.clicked.connect(self.handleFindCrystals)
        findCrystalsButton.setEnabled(False)
        self.showCrystalsCheckBox = QCheckBox("ShowCrystals", maximumHeight = 50)
        self.showCrystalsCheckBox.clicked.connect(self.handleShowCrystals)
        layout.addWidget(self.radioGroup, stretch=0)
        layout.addWidget(findCrystalsButton)
        layout.addWidget(self.showCrystalsCheckBox)
        
        self.setLayout(layout)
        
    def handleFindCrystals(self):
        self.findCrystals.emit()
        
    def handleShowCrystals(self):
        logger.debug(METHOD_ENTER_STR % 
                     self.showCrystalsCheckBox.isChecked())
        self.showCrystals.emit(self.showCrystalsCheckBox.isChecked())
        
        
        


STRING = 1
INT = 2
FLOAT = 3
LIST = 4

STATE_NONE = 0
STATE_LIST = 1

class Parser(object):
  def __init__(self, parser_info={}):
    self.parser_info = parser_info
    self.state = STATE_NONE

  def add_key_type(self, key, valtype):
    self.parser_info[key] = valtype

  def parse(self, lines):
    self.raw = []
    self.parsed = {}
    self.unknown_keys = []
    self.errors = []

    for line in lines:
      self.raw.append(line)
      stripped = line.strip()
      if stripped == '' or stripped[0] == '#':
        continue
      if self.state == STATE_LIST:
        # if indented, then part of list, otherwise 
        if line and line[0] in " \t":
          try:
            self.parsed[self.list_key].append(self.parse_val(self.list_type, line.strip()))
          except Exception as e:
            self.errors.append("Error parsing line '%s': %s"% (line.strip(), e))
          continue
        else:
          self.state = STATE_NONE

      bits = line.split(':', 1)
      key = bits[0].strip()
      if len(bits) > 1:
        val = bits[1].strip()
      else:
        val = ''

      if key in self.parser_info:
        val_type = self.parser_info[key]
        if type(val_type) == tuple and val_type[0] == LIST:
          self.parsed[key] = []
          self.state = STATE_LIST
          self.list_key = key 
          self.list_type = val_type[1]
          continue

        try:
          self.parsed[key] = self.parse_val(val_type, val)
        except Exception as e:
          self.errors.append("Error parsing line '%s': %s"% (line.strip(), e))
      else:
        self.unknown_keys.append(key)

    return self.parsed

  def parse_val(self, val_type, val):
    if callable(val_type):
      return val_type(val)
    elif val_type == STRING:
      return val
    elif val_type == INT:
      return int(val)
    elif val_type == FLOAT:
      return float(val)
    elif type(val_type) == tuple:
      bits = val.split(None,len(val_type)-1)
      return tuple([self.parse_val(bit_type, bit) for bit, bit_type in zip(bits, val_type)])

if __name__ == "__main__":
  # create a dictionary of keys and value types
  parser_info = {
      'name': STRING,
      'xtals': INT,
      'energy_range': (FLOAT, FLOAT),
      'filenames': (LIST, STRING),
      'exposure_files': (LIST, (FLOAT, STRING)),
      }
  p = Parser(parser_info)

  # some example data 
  raw = "name: foobar\nenergy_range: 7000.0 7140.0\nfilenames:\n  file1\n  file2\n  file3\nxtals: 10\nexposure_files:\n  7120.0 foo/what the/file1\n  7140.0 foo/what the/file2\nflub: hmm"

  # parse it
  p.parse(raw.split('\n'))

  # print it
  from pprint import pprint
  pprint(p.parsed)




import numpy as np
from numpy.linalg import norm

def Vector(x,y,z):
  return np.array([x,y,z])

Point = Vector

def angle_between(v1, v2):
  return np.arccos(np.dot(v1,v2) / norm(v1) / norm(v2))

class Plane(object):
  def __init__(self, p0, n):
    """
    Parameters:
      p0 - a point in the plane
      n  - unit vector normal to plane
    """
    self.p0 = np.array(p0)
    self.n = np.array(n)

  @classmethod
  def FromPoints(cls, p1, p2, p3):
    n = np.cross(p2-p1, p3-p1)
    n /= norm(n)

    return cls(p1, n)

  def __repr__(self):
    return "Plane(%s, %s)" % (self.p0.__repr__(), self.n.__repr__())

class Line(object):
  def __init__(self, l0, n):
    """
    Parameters:
      l0 - a point on the line
      n  - a unit vector in direction of lin
    """
    self.l0 = l0
    self.n = n

  @classmethod
  def FromPoints(cls, p1, p2):
    n = np.array(p2) - np.array(p1)
    n /= norm(n)

    return cls(p1, n)

  def __repr__(self):
    return "Plane(%s, %s)" % (self.p0.__repr__(), self.n.__repr__())

def intersect_line_with_plane(line, plane):
  a = np.dot(plane.p0 - line.l0, plane.n)
  b = np.dot(line.n, plane.n)  

  if (b == 0): # line parallel to plane
    if (a == 0): # in plane
      return plane.p0
    else: # out of plane 
      return None
  else:
    return line.l0 + a/b * line.n


def reflect_through_plane(point, plane):
  tmp = point - plane.p0
  return plane.p0 + tmp - 2 * np.dot(tmp, plane.n) * plane.n



from PyQt5.Qt import QApplication
import sys
import numpy as np

firstImage = np.zeros((1649,1475), dtype=np.int16)
#firstImage = np.zeros((1649,1475), dtype=np.float)
#firstImage = np.ones((1649,1475), dtype=np.float)
#firstImage = np.ones((1649,1475), dtype=np.int16)
from pyqtgraph import ImageView

app = QApplication(sys.argv)

imageView = ImageView()
print("type firstImage %s " % type(firstImage))
print("array dtype %s" % firstImage.dtype)
print ("firstImage.min/max %s %s" % (firstImage.min(), firstImage.max()) )
print (firstImage)
print(dir(firstImage))
print("here")
imageView.setImage(firstImage)
print("here")
imageView.show()

sys.exit(app.exec_())


'''
Created on May 17, 2019

@author: hammonds
'''
from PyQt5.QtWidgets import QApplication, QMainWindow, QSplitter, QTableView
import sys
from minixs.gui.qtcalibrator.model.exposurelistmodel import ExposureListModel
from minixs.gui.qtcalibrator.view.exposurelistview import ExposureListView

class CalibrationTestWindow(QMainWindow):
    
    def __init__(self, *args, **kwargs):
        super(CalibrationTestWindow, self).__init__( *args, **kwargs)
        self.splitter = QSplitter()
        
        self.tableView = ExposureListView()
        self.model = ExposureListModel()
        self.tableView.setModel(self.model)
        self.model.modelData.energies.append(3)
        self.splitter.addWidget(self.tableView)
        self.tableView2 = QTableView()
        self.tableView2.setModel(self.model)
        self.splitter.addWidget(self.tableView2)
        
        self.setCentralWidget(self.splitter)
        

if __name__ == '__main__':
    app = QApplication(sys.argv)
    mainWindow = CalibrationTestWindow()
    mainWindow.show()
    sys.exit(app.exec_())
    


"""
Calibration functions
"""

from minixs.exposure import Exposure
from minixs.emission import EmissionSpectrum, process_spectrum
from minixs import METHOD_ENTER_STR
try:
    from itertools import izip as zip     # Python 2 way, For python 3 can use zip directly
except ImportError:
    pass                                  # Nothing to do for python 3
from minixs.filter import get_filter_by_name
from minixs.constants import *
from minixs.gauss import gauss_leastsq, gauss_model
from minixs.parser import Parser, STRING, INT, FLOAT, LIST
from minixs.filetype import InvalidFileError
import logging

import numpy as np
logger = logging.getLogger(__name__)

def load(filename):
  """
  Load a calibration matrix from a file
  """
  c = Calibration()
  c.load(filename)
  return c

def find_maxima(pixels, direction, window_size = 3):
  """
  Find locations of local maxima of `pixels` array in `direction`.

  The location is calculated as the first moment in a window of half width
  `window_size` centered at a local maximum.

  Parameters
  ----------
  pixels : pixel array from an elastic exposure
  direction: minixs.DIRECTION_* indicating dispersive direction
  window_size : size in pixels around max for windowed average

  Returns
  -------
  xy : array of x, y and energy coordinates
  """

  # convert direction to axis (XXX make this a function call somewhere)
  rolldir = direction % 2
  logger.debug("direction, rolldir %s, %s" %(direction, rolldir))
  # shorten expressions below by aliasing pixels array
  p = pixels

  # build mask of local maxima locations
  local_max = np.logical_and(p >= np.roll(p,-1,rolldir), p > np.roll(p, 1, rolldir))

  # perform windowed averaging to find average col value in local peak
  colMoment = np.zeros(p.shape)
  norm = np.zeros(p.shape)

  # build vector of indices along column (row) 
  cols = np.arange(0,p.shape[rolldir])
  if rolldir == VERTICAL:
    cols.shape = (len(cols),1)

  # find first moments about local maxima
  for i in range(-window_size,window_size+1):
    logger.debug("Shape of cols, shape of p %s: %s" % (cols.shape, p.shape))
    colMoment += local_max * np.roll(cols * p, i, rolldir)
    norm += local_max * np.roll(p, i, rolldir)

  # calculate average
  windowedAvg = colMoment / norm
  windowedAvg[np.isnan(windowedAvg)] = 0

  # we only want the locations of actual maxima
  index = np.where(windowedAvg > 0)
 
  # pull out the pixel locations of the peak centers
  if rolldir == VERTICAL:
    y = windowedAvg[index]
    x = index[1]
  else:
    x = windowedAvg[index]
    y = index[0]

  # return N x 2 array of peak locations
  return np.vstack([x,y]).T


def find_combined_maxima(exposures, energies, direction, progress_updater=None):
  """
  Build array of all maxima locations and energies in a list of exposures

  Parameters
  ----------
    exposures: a list of Exposure objects
    energies:  a list of corresponding energies (must be same length as `exposures`)
    direction: the dispersive direction

  Returns
  -------
    Nx3 array with columns giving x,y,energy for each maximum
  """
  points = []

  numberOfExposures = len(exposures)
  exposureNumber = 0
  for exposure, energy in zip(exposures, energies):
    # extract locations of peaks
    logger.debug(exposure.filename)
    xy = find_maxima(exposure.pixels, direction)
    z = energy * np.ones((len(xy), 1))
    xyz = np.hstack([xy,z])
    points.append(xyz)
    if progress_updater is not None:
        exposureNumber += 1
        percentDone = float(exposureNumber * 100.)/numberOfExposures
        progress_updater(int(percentDone))

  return np.vstack(points)


FIT_QUADRATIC = 1
FIT_CUBIC   = 2
FIT_QUARTIC = 3
FIT_ELLIPSOID = 4

def fit_region(region, points, dest, fit_type = FIT_QUARTIC, return_fit=False):
  """
  Fit a smooth function to points that lie in region bounded by `region`

  Parameters
  ----------
    region - a rectangle defining the boundary of region to fit: [(x1,y1), (x2,y2)]
    points - an N x 3 array of data points
    dest - an array to store fit data in
    fit_type - type of fit to perform, ether FIT_QUADRATIC or FIT_CUBIC

  Returns
  -------
    Nothing

  The points array should contain three columns giving respectively x,y and z
  values of data points.  The x and y values should be between 0 and the width
  and height of `dest` respectively. They are in units of pixels, but may be
  real valued.  The z values can take any values.

  The entries in `points` with x,y coordinates falling within the bounds
  specified by `region` are fit to the model specified by `fit_type` using linear
  least squares. This model is then evaluated at all integral values of x and y
  in this range, with the result being stored in the corresponding location of
  `dest`.

  This is intended to be called for several different non-overlapping values of
  `region` with the same list of `points` and `dest`.

  Fit Types
  ---------
    FIT_QUADRATIC: z = Ax^2 + By^2 + Cxy + Dx + Ey + F
    FIT_CUBIC: z = Ax^3 + By^3 + Cx^2y + Dxy^2 + Ex^2 + Fy^2 + Gxy + Hx + Iy + J
  """
  logger.debug(METHOD_ENTER_STR)
  logger.debug("region %s" % region)
  logger.debug("points %s" % points)
#   for point in points:
#       print ("POINT_ARRAY %s" % point)
  logger.debug("dest %s" % dest)
  logger.debug("fit_type %s" %fit_type)
  logger.debug("return_fit %s" % return_fit)
  logger.debug("size of points %s" % ((points.shape),))
  logger.debug("Non zero values in points %s" % (np.nonzero(points),))
  # boundary coordinates
  (x1,y1),(x2,y2) = region

  # extract points inside this xtal region
  index = np.where(np.logical_and(
      np.logical_and(
        points[:,0] >= x1,
        points[:,0] < x2),
      np.logical_and(
        points[:,1] >= y1,
        points[:,1] < y2
        )
      ))
  x,y,z = points[index].T

  # if we have no points in this region, we can't fit anything
  # XXX this should pass the warning up to higher level code instead
  #     of printing it out to stdout
  if len(x) == 0:
    logger.warning ("No points in region: %s" % region)
    return

  # build points to evaluate fit at
  xxd, yyd = np.meshgrid(np.arange(x1,x2), np.arange(y1,y2))
  xxd = np.ravel(xxd)
  yyd = np.ravel(yyd)
  xx = xxd.astype('double')
  yy = yyd.astype('double')

  if fit_type == FIT_QUADRATIC:
    # fit to quadratic: z = Ax^2 + By^2 + Cxy + Dx + Ey + F
    A = np.vstack([x**2, y**2, x*y, x, y, np.ones(x.shape)]).T
    fit, r = np.linalg.lstsq(A,z)[0:2]

    # calculate residues
    rms_res = np.sqrt(r / len(z))[0]
    lin_res = sum(z - np.dot(A, fit)) / len(z)

    # evaluate at all points
    zz = np.dot(
           np.vstack([xx**2,yy**2,xx*yy,xx,yy,np.ones(xx.shape)]).T,
           fit
         ).T

  elif fit_type == FIT_CUBIC:
    # fit to cubic:
    #   Ax^3 + By^3 + Cx^2y + Dxy^2 + Ex^2 + Fy^2 + Gxy + Hx + Iy + J = z
    A = np.vstack([x**3,y**3,x**2*y,x*y**2,x**2, y**2, x*y, x, y, np.ones(x.shape)]).T
    fit, r = np.linalg.lstsq(A,z)[0:2]

    # calculate residues
    rms_res = np.sqrt(r / len(z))[0]
    lin_res = sum(z - np.dot(A, fit)) / len(z)

    # evaluate at all points
    zz = np.dot(
           np.vstack([
             xx**3, yy**3, xx**2*yy, xx*yy**2,
             xx**2, yy**2, xx*yy,
             xx, yy,
             np.ones(xx.shape)
           ]).T,
           fit
         ).T

  elif fit_type == FIT_QUARTIC:
    # fit to quartic:
    A = np.vstack([
      x**4,
      y**4,
      x**2 * y**2,
      # intentionally skip all terms with cubes
      x**2 * y,
      x * y**2,
      x**2,
      y**2,
      x * y,
      x,
      y,
      np.ones(x.shape)
      ]).T
    fit, r = np.linalg.lstsq(A,z)[0:2]

    # calculate residues
    rms_res = np.sqrt(r / len(z)) #[0]
    lin_res = sum(z - np.dot(A, fit)) / len(z)

    # evaluate at all points
    zz = np.dot(
           np.vstack([
             xx**4, yy**4, xx**2 * yy**2,
             xx**2 * yy, xx * yy**2,
             xx**2, yy**2, xx * yy,
             xx, yy,
             np.ones(xx.shape)
           ]).T,
           fit
         ).T

  elif fit_type == FIT_ELLIPSOID:
    raise Exception("Fit method not yet implemented.")
    # XXX this doesn't seem to work...
    # Fit to ellipsoid:
    # Ax^2 + By^2 + Cz^2 + Dxy + Eyz + Fzx + Gx + Hy + Iz = 1
    data = np.vstack([
      x*x,
      y*y,
      z*z,
      x*y,
      y*z,
      z*x,
      x,
      y,
      z
      ]).T

    w = np.ones(x.shape)
    fit, r = np.linalg.lstsq(data,w)[0:2]

    #rms_res = np.sqrt(r / len(w))[0]
    lin_res = (w - np.dot(data, fit)).sum() / len(w)
    rms_res = np.sqrt(((w - np.dot(data, fit))**2).sum() / len(w))

    # now solve for z in terms of x and y to evaluate
    A,B,C,D,E,F,G,H,I = fit
    a = C
    b = E*yy + F*xx + I
    c = A*xx**2 + B*yy**2 + D*xx*yy + G*xx + H*yy - 1
    zz = (-b + np.sqrt(b**2 - 4*a*c)) / (2*a)

  # fill the calibration matrix with values from fit
  dest[yyd,xxd] = zz

  if return_fit:
    return lin_res, rms_res, fit
  else:
    return lin_res, rms_res

def calibrate(filtered_exposures, energies, regions, 
              dispersive_direction, 
              fit_type=FIT_QUARTIC, 
              return_diagnostics=False,
              progress_updater=None):
  """
  Build calibration matrix from parameters in Calibration object

  Parameters
  ----------
    filtered_exposures: list of loaded and cleaned up Exposure objects
    energies: list of energies corresponding to exposures 
    regions: list of regions containing individual spectra
    dispersive_direction: direction of increasing energy on camera (minixs.const.{DOWN,UP,LEFT,RIGHT})

  Optional Parameters
  -------------------
    fit_type: type of fit (see fit_region() for more)
    return_diagnostics: whether to return extra information (residues and points used for fit)

  Returns
  -------
    calibration_matrix, [lin_res, rms_res, points]

    calibration_matrix: matrix of energies assigned to each pixel

    lin_res: average linear deviation of fit
    rms_res: avg root mean square residue of fit
    points: extracted maxima used for fit

    The last 3 of these are only returned if `return_diagnostics` is True.
  """
  # locate maxima
  logger.debug(METHOD_ENTER_STR)
  logger.debug(("Exposures %s\n, energies %s\nxtals %s\n" +
                 "dispersive_direction %s\nfit_type %s \n" + 
                 "return_diagnostics %s") % 
                 ( filtered_exposures, energies, regions,
                   dispersive_direction, fit_type,
                   return_diagnostics))
  points = find_combined_maxima(filtered_exposures, energies, 
                                dispersive_direction,
                                progress_updater=progress_updater)
  # create empty calibration matrix
  calibration_matrix = np.zeros(filtered_exposures[0].pixels.shape)

  # fit smooth shape for each crystal, storing fit residues
  lin_res = []
  rms_res = []
  fits = []
  regionNumber = 0
  numRegions = len(regions)
  progress_updater(0)
  for region in regions:
    lr, rr, fit = fit_region(region, points, calibration_matrix, fit_type, return_fit=True)
    lin_res.append(lr)
    rms_res.append(rr)
    fits.append(fit)
    if progress_updater is not None:
        regionNumber += 1
        percentDone = float(regionNumber * 100.)/numRegions
        progress_updater(int(percentDone))

  if return_diagnostics:
    return (calibration_matrix, points, (lin_res, rms_res, points, fits))
  else:
    return calibration_matrix, points

class Calibration(object):
  """
  A calibration matrix and all corresponding information
  """

  def __init__(self):
    logger.debug(METHOD_ENTER_STR)
    self.dataset_name = ""
    self.dispersive_direction = DOWN
    self.energies = []
    self.exposure_files = []
    self.filters = []
    self.xtals = []
    self.fit_points = []
    self.calibration_matrix = np.array([])

    self.filename = None

    self.load_errors = []

  def save(self, filename=None, header_only=False):
    if filename is None:
      filename = self.filename
    else:
      self.filename = filename

    with open(filename, 'w') as f:
      f.write("# miniXS calibration matrix\n#\n")
      f.write("# Dataset: %s\n" % self.dataset_name)
      f.write("# Dispersive Direction: %s\n" % DIRECTION_NAMES[self.dispersive_direction])
      f.write("#\n")
      f.write("# Energies and Exposures:\n")
      for en, ex in zip(self.energies, self.exposure_files):
        f.write("#   %5.2f %s\n" % (en,ex))
      f.write("#\n")

      f.write("# Filters:\n")
      for fltr in self.filters:
        f.write('#   %s: %s\n' % (fltr.name, fltr.get_str()))
      f.write("#\n")

      f.write("# Xtal Boundaries:\n")
      for (x1,y1), (x2,y2) in self.xtals:
        f.write("#   %3d %3d %3d %3d\n" % (x1,y1,x2,y2))
      f.write("#\n")

      if (not header_only and
          self.calibration_matrix is not None and
          len(self.calibration_matrix) > 0 and 
          len(self.calibration_matrix.shape) == 2):
        f.write("# %d x %d matrix follows\n" % self.calibration_matrix.shape)

        np.savetxt(f, self.calibration_matrix, fmt='%.3f')

  def load(self, filename=None, header_only=False):
    """
    Load calibration information from saved file

    Parameters
    ----------
      filename: name of file to load
      header_only: whether to load only the header or full data

    Returns
    -------
      True if load was successful
      False if load encountered an error

      Error messages are stored as strings in the list `Calibration.load_errors`.
    """
    logger.debug(METHOD_ENTER_STR)
    self.load_errors = []

    if filename is None:
      filename = self.filename
    else:
      self.filename = filename

    parser = Parser({
      'Dataset': STRING,
      'Dispersive Direction': STRING,
      'Energies and Exposures': (LIST, (FLOAT, STRING)),
      'Filters': (LIST, STRING),
      'Xtal Boundaries': (LIST, (INT, INT, INT, INT)),
      })

    header = []
    with open(filename, 'r') as f:
      line = f.readline()
      #XXX this should be done elsewhere...
      if line.strip().lower() != '# minixs calibration matrix':
        raise InvalidFileError()

      pos = f.tell()
      line = f.readline()
      while line:
        if line[0] == "#":
          header.append(line[2:])
        elif header_only:
          break
        else:
          f.seek(pos)
          self.calibration_matrix = np.loadtxt(f)
          if len(self.calibration_matrix.shape) == 1:
            self.calibration_matrix.shape = (1,self.spectrum.shape[0])
          break

        pos = f.tell()
        line = f.readline()

    # parse
    parsed = parser.parse(header)
    if parser.errors:
      self.load_errors += parser.errors

    self.dataset = parsed.get('Dataset', '')

    # check dispersive direction
    dirname = parsed.get('Dispersive Direction', DOWN)
    if dirname in DIRECTION_NAMES:
      self.dispersive_direction = DIRECTION_NAMES.index(dirname)
    else:
      self.load_errors.append("Unknown Dispersive Direction: '%s', using default." % dirname)

    # split up energies and exposure files
    key = 'Energies and Exposures'
    if key in parsed.keys():
      self.energies = [ee[0] for ee in parsed[key]]
      self.exposure_files= [ee[1] for ee in parsed[key]]

    # read in filters
    for filter_line in parsed.get('Filters', []):
      name,val = filter_line.split(':')
      name = name.strip()
      fltr = get_filter_by_name(name)
      if fltr == None:
        self.load_errors.append("Unknown Filter: '%s' (Ignoring)" % name)
      else:
        fltr.set_str(val.strip())
        self.filters.append(fltr)

    self.xtals = [
        [[x1,y1],[x2,y2]]
        for x1,y1,x2,y2 in parsed.get('Xtal Boundaries', [])
        ]

    return len(self.load_errors) == 0

  def calibrate(self, fit_type=FIT_QUARTIC, progressUpdater=None):
    # load exposure files
    exposures = [Exposure(f) for f in self.exposure_files]
    
    # apply filters
    for exposure, energy in zip(exposures, self.energies):
      for f in self.filters:
        f.filter(exposure.pixels, energy)

    # calibrate
    self.calibration_matrix, self.fit_points, diagnostics = calibrate(exposures,
                                                     self.energies,
                                                     self.xtals,
                                                     self.dispersive_direction,
                                                     fit_type,
                                                     return_diagnostics=True,
                                                     progress_updater=progressUpdater)

    # store diagnostic info
    self.lin_res, self.rms_res, self.fit_points, self.fits = diagnostics

  def xtal_mask(self):
    """
    Generate a mask of the xtal regions

    Returns
    -------
      An binary array of the same shape as the calibration matrix with 0's outside of xtal regions and 1's inside
    """

    if self.calibration_matrix is None:
      return None

    mask = np.zeros(self.calibration_matrix.shape, dtype=np.bool)
    for (x1,y1),(x2,y2) in self.xtals:
      mask[y1:y2,x1:x2] = 1

    return mask

  def energy_range(self):
    """
    Find min and max energies in calibration matrix

    Returns
    -------
      (min_energy, max_energy)
    """
    return (self.calibration_matrix[np.where(self.calibration_matrix > 0)].min(), self.calibration_matrix.max())

  def diagnose(self, return_spectra=False, filters=None):
    """
    Process all calibration exposures and fit to gaussians, returning parameters of fit

    Parameters
    ----------
    return_spectra: whether to return processed calibration spectra

    Returns
    -------
    (diagnostics, [processed_spectra])

    diagnostics: an array with one row for each calibration exposure
                 the columns are:
                   incident beam energy
                   amplitude
                   E0
                   sigma

                 the best Gaussian fit to the data is given by:
                   exp(-(E-E0)**2/(2*sigma**2))

    if `return_spectra` is True, then a list of XES spectra will be returned (one for each calibration exposure)
    """

    emin, emax = self.energy_range()
    emission_energies = np.arange(emin, emax, .2)

    diagnostics = np.zeros((len(self.energies), 4))

    if return_spectra:
      spectra = []

    for i in range(len(self.energies)):
      if i % 10 == 0:
        print (i)
      energy = self.energies[i]
      exposure = Exposure(self.exposure_files[i])
      if filters is not None:
        exposure.apply_filters(energy, filters)

      s = process_spectrum(self.calibration_matrix, exposure, emission_energies, 1, self.dispersive_direction, self.xtals)
      x = s[:,0]
      y = s[:,1]

      fit, ier = gauss_leastsq((x,y), (y.max(), energy, 1.0))

      if not (0 < ier < 5):
        continue

      diagnostics[i,0] = energy
      diagnostics[i,1:] = fit

      if return_spectra:
        xes = EmissionSpectrum()
        xes.incident_energy = energy
        xes.exposure_files = [exposure.filename]
        xes.set_spectrum(s)
        spectra.append(xes)

    diagnostics = diagnostics[np.where(diagnostics[:,0] != 0)]

    if return_spectra:
      return (diagnostics, spectra)
    else:
      return diagnostics

  def loadFitPoints(self):
    logger.debug(METHOD_ENTER_STR)
    exposures = [Exposure(f) for f in self.exposure_files]
    
    # apply filters
    for exposure, energy in zip(exposures, self.energies):
          for f in self.filters:
            f.filter(exposure.pixels, energy)
    logger.debug("Exposures, energies %s, %s" % (exposures, self.energies))
    points = find_combined_maxima(exposures, self.energies, self.dispersive_direction)
    logger.debug("Points %s" % points)
    self.fit_points = points


'''
 Copyright (c) 2019, UChicago Argonne, LLC
 See LICENSE file.
'''
import logging
from os.path import realpath
import traceback

from PyQt5.Qt import pyqtSlot, QFileDialog, QProgressBar
from PyQt5.QtWidgets import QAction, QMessageBox
from pyqtgraph.graphicsItems.ViewBox.ViewBox import ViewBox

import PyQt5.QtCore as qtCore
import PyQt5.QtWidgets as qtWidgets
from minixs import METHOD_ENTER_STR, filetype
from minixs.calibrate import Calibration
from minixs.config.minixsconfigparser import MinixsConfigParser, FILE_SECTION, \
    LAST_DIRECTORY_STR
from minixs.exposure import Exposure
from minixs.filetype import InvalidFileError
from minixs.filter import get_filter_by_name
from minixs.gui import util
from minixs.gui.qtcalibrator.ImageViewQt import ImageViewQt
from minixs.gui.qtcalibrator.calibratorexposureselector import CalibratorExposureSelector
from minixs.gui.qthelpers.wildcards import WILDCARD_CALIB, WILDCARD_XTAL, \
    WILDCARD_XTAL_EXPORT
from minixs.misc import find_xtal_boundaries
import numpy as np
from PyQt5.QtCore import QCoreApplication

try:
    from itertools import izip as zip     # Python 2 way, For python 3 can use zip directly
except ImportError:
    pass                                  # Nothing to do for python 3
logger = logging.getLogger(__name__)

APP_NAME = "Calibrator"
SAVE_CALIB_TITLE = "Save Calibration File"
OPEN_CALIB_TITLE = "Open Calibration File"
EXPORT_XTAL_TITLE = "Export Crystal Boundaries"
IMPORT_XTAL_TITLE = "Import Crystal Boundaries"
INVALID_FILE_TEXT = "Invalid file"
INVALID_FILE_TYPE_TEXT = "Invalid file type"


class CalibrationModel(Calibration):
    def __init__(self):
        super(CalibrationModel, self).__init__()

class CalibratorMainWindow(qtWidgets.QMainWindow):
    '''
    Defines the main window for the Calibrator
    '''


    def __init__(self, parent=None):
        '''
        Constructor
        '''
        super(CalibratorMainWindow, self).__init__(parent)
        
        self.model = CalibrationModel()
        self.config = MinixsConfigParser()
        self.createMenuBar()
        
        self.splitter = qtWidgets.QSplitter(qtCore.Qt.Vertical)
        self.exposureSelector = CalibratorExposureSelector()
        self.exposureSelector.setCalibrationModel(self.model)
        self.exposureSelector.exposureListChanged.connect(self.handleExposureListChanged)
        self.exposureSelector.energiesListChanged.connect(self.handleEnergiesListChanged)
        self.exposureSelector.tableAppendRow.connect(self.handleTableAppendRow)
        self.exposureSelector.tableDeleteRows[list].connect(self.handleTableDeleteRows)
        self.exposureSelector.clearEnergies.connect(self.handleClearEnergies)
        self.exposureSelector.clearExposures.connect(self.handleClearExposures)
        self.exposureSelector.tableView.singleEnergyChanged[int, float].connect(self.handleSingleEnergyChanged)
        self.exposureSelector.tableView.singleExposureNameChanged[int, str].connect(self.handleSingleExposureNameChanged)
        self.splitter.addWidget(self.exposureSelector)
        self.imageView = ImageViewQt()
        self.imageView.findCrystals.connect(self.handleFindCrystals)
        self.imageView.calibrateButton.clicked.connect(self.handleCalibrate)
        self.imageView.filters.filterChanged.connect(self.handleFilterChanged)
        self.imageView.filters.dispersiveDirection.currentIndexChanged[int].connect(self.handleDispersiveDirectionChanged)
        self.splitter.addWidget(self.imageView)
        
        self.setCentralWidget(self.splitter)
        self.setWindowTitle(APP_NAME)
        statusBar = self.statusBar()
        self.progressBar = QProgressBar()
        self.progressBar.setMinimum(0)
        self.progressBar.setMaximum(100)
        self.progressBar.setMaximumSize(300, 20)
        statusBar.addPermanentWidget(self.progressBar, stretch = 25)
        self.changed = False
        self.calibrationValid(False)

    def Changed(self, changed=True):
        self.changed = changed
        
    def calibrationValid(self, valid):
        self.calibration_valid = valid
        self.imageView.calibrateButton.setEnabled(not valid)
        
    def createMenuBar(self):
        '''
        '''
        mainMenu = self.menuBar()
        fileMenu = mainMenu.addMenu('&File')
        viewMenu = mainMenu.addMenu('&View')
        
        loadCalibAction = QAction("&Open...", self)
        saveCalibAction = QAction("&Save...", self)
        importXtalAction = QAction("&Import Crystals...", self)
        exportXtalAction = QAction("&Export Crystals...", self)
        fileMenu.addAction(loadCalibAction)
        fileMenu.addAction(saveCalibAction)
        fileMenu.addSeparator()
        fileMenu.addAction(importXtalAction)
        fileMenu.addAction(exportXtalAction)
        viewFitPointsAction = QAction("&Fit Points", self, checkable=True)
        viewMenu.addAction(viewFitPointsAction)
        loadCalibAction.triggered.connect(self.loadCalibration)
        saveCalibAction.triggered.connect(self.saveCalibration)
        importXtalAction.triggered.connect(self.importXtals)
        exportXtalAction.triggered.connect(self.exportXtals)
        viewFitPointsAction.triggered.connect(self.handleViewFitPoints)

    @pyqtSlot()
    def exportXtals(self):
        logger.debug(METHOD_ENTER_STR)
        lastPath = self.config.getLastFilePath()
        filename, _ = QFileDialog.getSaveFileName(self, EXPORT_XTAL_TITLE,
                                                   lastPath,
                                                   filter=WILDCARD_XTAL_EXPORT)
        self.viewToModel()
        if (filename != "") and (filename is not None):
            with open(filename, 'w') as f:
                f.write("# miniXS crystal boundaries\n")
                f.write("# x1\ty1\tx2\ty2\n")
                for (x1,y1),(x2,y2) in self.model.xtals:
                    f.write("%d\t%d\t%d\t%d\n" % (x1,y1,x2,y2))
        
#     def handleExposuresDeleted(self):
#         self.setButtonsNoExposures()
        
    def handleCalibrate(self):
        '''
        Activate the calibration method to add the calibration to the model.  
        Check and make sure we have what we need like havig images and defined
        ROIs where calibration should be calculated.
        '''
        logger.debug(METHOD_ENTER_STR)
        self.progressBar.setValue(0)
        self.viewToModel()
        valid, errors = self.validate()
        if not valid:
            errors = [ '% 2d) %s' % (i+1, err) for i, err in enumerate(errors) ]
            message = 'The following must be fixed before calibrating:\n\n' + '\n\n'.join(errors)
            QMessageBox.warning(self, "Calibration Warning", message)
            self.calibration_invalid = True
            return
        self.statusBar().showMessage('Calibrating...  Please Wait')
        
        try:
            self.model.calibrate(progressUpdater=self.progress_updater)
        except Exception as ex:
            tb = traceback.extract_stack()
            tb_string = ""
            logger.debug(tb)
            for f, line, m, stm in tb:
                part = "file %s,  Line: %d, method: %s\n statement: %s\n" % (f, line, m, stm)
                logger.debug("part %s" % part)
                tb_string.join(part)
            logger.debug("tb_string %s" % tb_string)
            QMessageBox.critical(self, "Error", "Error While calibrating\n %s %s" % (ex, tb_string))
            logger.exception(ex)
            return
        self.imageView.matrixView.setImage(self.model.calibration_matrix)
        logger.info("rms_res %s, lin_res %s" % (self.model.rms_res, self.model.lin_res))
        self.imageView.tools.calibrationRadioButton.setChecked(True)
        self.statusBar().showMessage("Done Calibrating")
        self.calibrationValid(True)
        self.Changed()
    
    @pyqtSlot(int)
    def handleDispersiveDirectionChanged(self, direction):
        self.model.dispersive_direction = direction
        self.calibrationValid(False)
        self.Changed()
    
    def handleExposureListChanged(self):
        logger.debug(METHOD_ENTER_STR)
        valid, exposureEnergies, exposureNames = self.exposureSelector.getData()
        self.imageView.loadImages(exposureNames, exposureEnergies)
        logger.debug("Filenames %s" % ((exposureNames, exposureEnergies),))
        self.calibrationValid(False)
        self.Changed()
        
    def handleEnergiesListChanged(self):
        logger.debug(METHOD_ENTER_STR)
        self.calibrationValid(False)
        self.Changed()
        
    def handleFilterChanged(self):
        self.calibrationValid(False)
        
    def handleFindCrystals(self):
        logger.debug(METHOD_ENTER_STR)
        self.viewToModel()
        
        if not self.model.exposure_files:
            return
        
        self.statusBar().showMessage("Finding Boundaries... Please Wait")
        
        exposures = [Exposure(f) for f in self.model.exposure_files]
        for exposure, energy in zip(exposures, self.model.energies):
            exposure.apply_filters(energy, self.model.filters)
            
        xtals = find_xtal_boundaries(exposures)
        
        logger.debug("Crystal boundaries %s" % ((xtals,)))
        
        if xtals is None:
            message = 'Unable to determine boundaries. It may help to increase the low cutoff value.'
            errdlg = QMessageBox.error(self, "Error", message)
            errdlg.ShowModal()
            errdlg.Destroy()
        else:
            self.model.xtals = xtals
            self.imageView.tools.rects = self.model.xtals
#           self.view.imageView.Refresh()
            self.Changed()
            self.calibrationValid(False)
        self.imageView.imageView.addROIs(xtals)

    def handleShowCrystals(self):
        pass

    def handleTableAppendRow(self):
        logger.debug(METHOD_ENTER_STR)
        
    def handleTableDeleteRows(self, rows):
        logger.debug(METHOD_ENTER_STR % (rows,))
        self.imageView.deleteImages(rows)
        valid, exposureEnergies, exposureNames = self.exposureSelector.getData()
        self.imageView.loadImages(exposureNames, exposureEnergies)
        logger.debug("Filenames %s" % ((exposureNames, exposureEnergies),))
        self.calibrationValid(False)
        self.Changed()
        
    def handleClearEnergies(self):
        logger.debug(METHOD_ENTER_STR)
        
    def handleClearExposures(self):
        logger.debug(METHOD_ENTER_STR)
        self.imageView.imageView.deleteAllImages()
    
    def handleSingleEnergyChanged(self, row, energy):
        logger.debug(METHOD_ENTER_STR  % ((row, energy),))
        valid, exposureEnergies, exposureNames = self.exposureSelector.getData()
        self.imageView.loadImages(exposureNames, exposureEnergies)
        logger.debug("Filenames %s" % ((exposureNames, exposureEnergies),))
        self.calibrationValid(False)
        self.Changed()
        
    def handleSingleExposureNameChanged(self, row, exposureName):
        logger.debug(METHOD_ENTER_STR  % ((row, exposureName),))
        valid, exposureEnergies, exposureNames = self.exposureSelector.getData()
        self.imageView.loadImages(exposureNames, exposureEnergies)
        logger.debug("Filenames %s" % ((exposureNames, exposureEnergies),))
        self.calibrationValid(False)
        self.Changed()

    
    @pyqtSlot(bool)
    def handleViewFitPoints(self, checkState=True):
        logger.debug(METHOD_ENTER_STR, checkState)
        self.viewToModel()
        if checkState:
            self.imageView.imageView.viewBox.addItem( self.imageView.imageView.fitPointsView)
            self.imageView.imageView.fitPointsView.setVisible(True)
            self.model.loadFitPoints()
            logger.debug("fit_points %s" % self.model.fit_points)
            self.imageView.imageView.fitPointsView.setData(x=self.model.fit_points[:,0], y=self.model.fit_points[:,1])
            
        else:
            self.imageView.imageView.viewBox.removeItem(self.imageView.imageView.fitPointsView)
            self.imageView.imageView.fitPointsView.setVisible(False)
        
    @pyqtSlot()
    def loadCalibration(self):
        '''
        Load a saved calibration from file
        '''
        logger.debug(METHOD_ENTER_STR)
        lastFilePath = self.config.getLastFilePath()
        filename, _ = QFileDialog.getOpenFileName(self, OPEN_CALIB_TITLE, 
                                                  lastFilePath,
                                                  filter=WILDCARD_CALIB)
        filename = str(filename)
        if filename != "" and (not filename is None):
            try:
                success = self.model.load(filename)
                if not success:
                    errDlg = QMessageBox()
                    errDlg.setText(("Warning: some errors were encountered while " +
                                    " loading the calibration file:\n\n  %s") \
                                     % '\n  '.join(self.model.load_errors))
                    errDlg.setIcon(QMessageBox.Critical)
                    errDlg.setStandardButtons(QMessageBox.Cancel | QMessageBox.Ok)
                    
                    retVal = errDlg.exec_()
                    return
            except InvalidFileError as ex:
                errDlg = QMessageBox()
                errDlg.setText(("Warning: some errors were encountered while " +
                                " loading the calibration file:\n\n  %s") \
                                 % '\n  '.join(self.model.load_errors))
                errDlg.setIcon(QMessageBox.Critical)
                errDlg.setStandardButtons(QMessageBox.Cancel | QMessageBox.Ok)
                
                retVal = errDlg.exec_()
                return
        else:
            return
                
        self.modelToView()
        
        
    @pyqtSlot()
    def importXtals(self):
        logger.debug(METHOD_ENTER_STR)
        lastPath = self.config.getLastFilePath()
        filename, _ = QFileDialog().getOpenFileName(self, IMPORT_XTAL_TITLE, 
                                                    lastPath,
                                                    filter=WILDCARD_XTAL)
        if filename != "" and filename is not None:
            error = None
            t = filetype.determine_filetype(filename)
            if t == filetype.FILE_CALIBRATION:
                ci = Calibration()
                ci.load(filename, header_only=True)
                self.model.xtals = ci.xtals
            elif t == filetype.FILE_XTALS or t == filetype.FILE_UNKNOWN:
                data = None
                try:
                    data = np.loadtxt(filename)
                    self.model.xtals = [[[x1,y1],[x2,y2]] for x1, y1, x2, y2 in data]
                except ValueError:
                    if data is not None:
                        if len(data.shape) == 1:
                            num = data.shape[0]
                        else:
                            num = data.shape1
                        if num !=4:
                            error = "Crystal file must contain 4 columns"
                        else:
                            error = INVALID_FILE_TEXT
                            
                    else:
                        error = INVALID_FILE_TEXT
            else:
                error = INVALID_FILE_TYPE_TEXT
                
            if error:
                QMessageBox.critical(self, "Error", "Invalid Xtal File")
            self.imageView.imageView.replaceXtals(self.model.xtals)
            self.calibrationValid(False)
            self.Changed()
        
    def modelToView(self):
        logger.debug(METHOD_ENTER_STR)
        self.exposureSelector.datasetNameTxt.setText(self.model.dataset_name)
        self.imageView.filters.dispersiveDirection.setCurrentIndex(self.model.dispersive_direction)
        
        self.exposureSelector.clearExposuresButton.click()
        self.exposureSelector.clearEnergiesButton.click()
        
        self.exposureSelector.appendRows(self.model.energies, self.model.exposure_files)
        
        self.imageView.filters.setFilters(self.model.filters)
        
        self.imageView.imageView.addROIs(self.model.xtals)
        self.imageView.matrixView.setImage(self.model.calibration_matrix)
        
    def progress_updater(self, progressPercent=0):
        logger.debug(METHOD_ENTER_STR % progressPercent)
        self.progressBar.setValue(progressPercent)
        QCoreApplication.processEvents()
        
        
    def validate(self):
        errors = []
        valid = True
    
        if not self.exposure_list_valid:
            valid = False
            errors.append("The exposure list is invalid. Make sure that each row contains an energy and an exposure filename.")
    
        if len(self.model.energies) < 2:
            valid = False
            errors.append("At least two calibration exposures are required for calibration. A larger number of exposures will give a better fit.")
    
        if len(self.model.xtals) < 1:
            valid = False
            errors.append("Define the boundary of at least one crystal.")
    
        intersecting = False
        for xa in self.model.xtals:
            if intersecting:
                break
    
            for xb in self.model.xtals:
                if xa == xb:
                    continue
                if util.xtals_intersect(xa, xb):
                    valid = False
                    errors.append("Crystal boundaries may not overlap.")
                    intersecting = True
                    break
    
        return valid, errors

    @pyqtSlot()
    def saveCalibration(self):
        logger.debug(METHOD_ENTER_STR)
        header_only = False
        if self.calibration_valid == False:
            errDlg = QMessageBox()
            errDlg.setIcon(QMessageBox.Warning)
            errDlg.setText( 
                    "Warning: You have changed parameters since last " +
                    "calibrating. Saving now will only save the " +
                    "parameters, and not the matrix itself.")
            errDlg.StandardButtons(QMessageBox.Cancel | QMessageBox.Ok)
            
            retVal = errDlg.exec_()
            if retVal == QMessageBox.Ok:
                logger.debug("Warning box passed OK, saving header only")
                header_only=True
            else:
                return
        lastPath = self.config.getLastFilePath()
        filename, _ = QFileDialog.getSaveFileName(self, SAVE_CALIB_TITLE, 
                                                  lastPath,
                                                  filter=WILDCARD_CALIB )

        if (filename != "") and (not filename is None):
            self.viewToModel()
            self.model.save(filename, header_only=header_only)
        #save last used directory back to the config
        self.config.set(FILE_SECTION, LAST_DIRECTORY_STR, realpath(filename))
        self.config.updateFile()
            
    def viewToModel(self):
        self.model.dataset_name = str(self.exposureSelector.datasetNameTxt.text())
        self.model.dispersive_direction = self.imageView.filters.dispersiveDirection.currentIndex()
        logger.debug("Dispersive Direction %s " % self.model.dispersive_direction)
        # get Energies and exposures
        valid, energies, exposure_files = self.exposureSelector.getData()
        self.exposure_list_valid = valid
        self.model.energies = energies
        self.model.exposure_files = exposure_files
        
        self.model.filters = []
        filters = self.imageView.filters.getFilters()
        for (name, enabled, val) in filters:
            if enabled:
                fltr = get_filter_by_name(name)
                fltr.set_val(val)
                self.model.filters.append(fltr)
        if len(self.imageView.imageView.rois) >0:
            dir (self.imageView.imageView.rois[0].pos())        
        self.model.xtals = [
            [[int(round(roi.pos().x())), int(round(roi.pos().y()))], \
             [int(round(roi.pos().x() + roi.size().x())), int(round(roi.pos().y() + roi.size().y()))]] \
                 for roi in self.imageView.imageView.rois
            ]
     



"""
Miscellanous functions
"""

import numpy as np

try:
    from itertools import izip as zip     # Python 2 way, For python 3 can use zip directly
except ImportError:
    pass                                  # Nothing to do for python 3
from itertools import product as iproduct
from minixs.scanfile import ScanFile
import logging
logger=logging.getLogger(__name__)

def gen_rects(horizontal_bounds=None, vertical_bounds=None):
  """Convert lists of horizontal and vertical boundary locations into rectangles"""
  return [
      ((l,t),(r,b))
      for (l,r) in zip(horizontal_bounds[0:-1], horizontal_bounds[1:])
      for (t,b) in zip(vertical_bounds[0:-1], vertical_bounds[1:])
      ]

def gen_file_list(prefix, vals, zero_pad=3, suffix='.tif'):
  """Generate a list of numerically sequenced file names"""
  fmt = '%s%%0%dd%s' % (prefix, zero_pad, suffix)
  return [fmt % val for val in vals]

def read_scan_info(scanfile, columns):
  """Read a subset of columns from a scan file

  The default columns correspond to the typical Mono Energy an I0 columns.
  The returned list is transposed so that the following works:

  c0, c6 = read_scan_info("scanfile.0001", [0,6])
  """

  s = ScanFile(scanfile)
  return s.data[:,columns].transpose()

def determine_dispersive_direction(e1, e2, threshold=.75, sep=20):
  """Given two exposures with increasing energy, determine
  the dispersive direction on the camera"""

  p1 = e1.pixels.copy()
  p2 = e2.pixels.copy()

  p1[np.where(p1 > 10000)] = 0
  p2[np.where(p2 > 10000)] = 0

  tests = [
      (0, (1,sep)), # UP
      (1, (-sep,-1)), # LEFT
      (0, (-sep,-1)),   # DOWN
      (1, (1,sep))    # RIGHT
      ]

  diffs = [
      np.min([ np.sum(np.abs(p2 - np.roll(p1,i,axis)))
            for i in np.arange(i1,i2) ])
      for axis, (i1,i2) in tests ]

  if np.min(diffs) / np.average(diffs) < threshold:
    return np.argmin(diffs)

  return -1

def build_rixs(spectra, energies):
  total_length = np.sum([len(s) for s in spectra])
  full_spectrum = np.zeros((total_length, 6))

  i = 0
  for s,E in zip(spectra, energies):
    full_spectrum[i:i+len(s),0] = E
    full_spectrum[i:i+len(s),1:] = s
    i += len(s)

  return full_spectrum

def save_rixs(filename, rixs):
  with open(filename, 'w') as f:
    f.write("#    E_incident      E_emission       Intensity           Sigma          Counts      Num_pixels\n")

    fmt = ('% 15.2f', '% 15.2f', '% 15.8e','% 15.8e','% 15d', '% 15d')
    savetxt(f, rixs, fmt=fmt)


def rixs_xes_cut(rixs, energy):
  energies = np.unique(rixs[:,0])
 
  i = argmin(np.abs(energies - energy))

  return rixs[np.where(rixs[:,0] == energies[i])]

def rixs_pfy_cut(rixs, energy):
  energies = np.unique(rixs[:,1])
 
  i = np.argmin(np.abs(energies - energy))

  return rixs[np.where(rixs[:,1] == energies[i])]


def rixs2d(rixs):
  inc_energies = np.unique(rixs[:,0])
  emit_energies = np.unique(rixs[:,1])

  rixs2d = np.zeros((len(emit_energies), len(inc_energies)))

  i = len(inc_energies)
  return rixs[:,2].reshape((i,len(rixs)/i)).T

def plot_rixs_contour(rixs, plot_log=False, aspect=1):
  incE = np.unique(rixs[:,0])
  emitE = np.unique(rixs[:,1])
  r2d = rixs2d(rixs)
  if plot_log:
    r2d = np.log10(r2d)

  from matplotlib.pyplot import contourf, figure, colorbar
  figure()
  contourf(incE, emitE, r2d, 50, aspect=aspect)
  colorbar()

def plot_rixs(rixs, start=0, end=-1, plot_log=False, aspect=1):
  incE = np.unique(rixs[:,0])
  emitE = np.unique(rixs[:,1])
  r2d = rixs2d(rixs)
  if plot_log:
    r2d = np.log10(r2d)

  from matplotlib.pyplot import imshow, figure
  figure()
  imshow(
      r2d[:,start:end],
      extent=(incE[start],incE[end],emitE[-1],emitE[0]),
      aspect=aspect)

def plot_spectrum(s, **kwargs):
  plot_errorbars = True
  if kwargs.has_key('errorbar'):
    plot_errorbars = kwargs['errorbar']
    del(kwargs['errorbar'])

  new_plot = True
  if kwargs.has_key('new_plot'):
    new_plot = kwargs['new_plot']
    del(kwargs['new_plot'])

  normalize = False
  if kwargs.has_key('normalize'):
    normalize = kwargs['normalize']
    del(kwargs['normalize'])


  from matplotlib.pyplot import plot, errorbar, figure

  if new_plot:
    figure()

 
  n = 1
  if normalize:
    n = np.sum(s[:,1]) / len(s[:,1])

  if plot_errorbars:
    errorbar(s[:,0], s[:,1]/n, s[:,2]/n, **kwargs)
  else:
    plot(s[:,0], s[:,1]/n, **kwargs)

def _find_boundaries(p, axis):
  """
  Helper function for find_xtal_boundaries
  """
  flat = p.sum(1-axis)

  # find boundaries
  mask = 1 * (flat > p.shape[1-axis])
  diffs = np.diff(mask)
  b1 = list(np.where(diffs==1)[0])
  b2 = list(np.where(diffs==-1)[0])

  # if nothing was found, xtals cover entire face
  if len(b1) == 0:
    b1 = [0]
  if len(b2) == 0:
    b2 = [p.shape[axis]]
  
  if b1[0] > b2[0]:
    b1.insert(0,0)
  if b2[-1] < b1[-1]:
    b2.append(p.shape[axis])

  return b1,b2

def find_xtal_boundaries(filtered_exposures, shrink=1):
  """
  From a full set of filtered calibration exposures, determine crystal boundaries

  Parameters
  ----------
    filtered_exposures: a list of Exposures
    shrink: the number of pixels to shrink determined boundary (default 1)

  The `shrink` parameter can be used to avoid edge effects or slight aparallelism.
  """
  # combine exposures into one
  p = filtered_exposures[0].pixels.copy()
  for e in filtered_exposures[1:]:
    np.add(p,e.pixels,p)

  x1s,x2s = _find_boundaries(p, 1)
  y1s,y2s = _find_boundaries(p, 0)
#   y1s = [2]
#   y2s = [193]


  if len(x1s) != len(x2s) or len(y1s) != len(y2s):
    return None

  xtals = []
  for x1,x2 in zip(x1s, x2s):
    for y1,y2 in zip(y1s, y2s):
      xtals.append([[x1+shrink,y1+shrink],[x2-shrink,y2-shrink]])

  return xtals

def find_edges(pixels, direction=0, sign=-1, thresh=5):
  """
  Find edges of regions with pixels above thresh
  """
  return np.where(np.diff((pixels.max(direction)>thresh)*1) == sign)[direction]

def find_xtal_regions2(filtered_exposures, direction, thresh):
  r = [find_edges(e.pixels, direction, -1, thresh) for e in filtered_exposures]
  print (r)
  l = [find_edges(e.pixels, direction, +1, thresh) for e in filtered_exposures]

  return l,r

def collection_angle_correction(ci, d, return_theta=False):
  """
  Calculate angular spread in dispersive direction (in radians) for each pixel

  This should by used to correct for the fact that different pixels bin different energy widths (or alternatively cover different solid angles).

  This code has not been fully verified, so do not use at this time.
 
  Parameters
  ----------
    ci: Calibration object containing calibration matrix and xtals
    d:  lattice spacing
    return_theta: if True, return a matrix of angle values for each pixel
  """
  cac =  np.zeros(ci.calibration_matrix.shape)
  if return_theta:
    fulltheta =  np.zeros(ci.calibration_matrix.shape)

  E0 = np.pi * 1973.26 / d

  for xtal in ci.xtals:
    (x1,y1),(x2,y2) = xtal
    E = ci.calibration_matrix[y1:y2,x1:x2]

    dE = (np.roll(E,-1,0) - np.roll(E,1,0)) / 2.
    # fix up end points by extrapolation
    dE[0] = 2*dE[1] - dE[2]
    dE[-1] = 2*dE[-2] - dE[-3]

    theta = np.arccos(E0 / E)
    dtheta = dE / E / np.tan(theta)
 
    if return_theta:
      fulltheta[y1:y2, x1:x2] = theta
    cac[y1:y2, x1:x2] = dtheta

  if return_theta:
    return fulltheta, cac
  else:
    return cac



'''
 Copyright (c) 2019, UChicago Argonne, LLC
 See LICENSE file.
'''
import sys
import unittest
from PyQt5.QtWidgets import QApplication, QTableWidgetSelectionRange
from PyQt5.QtTest import QTest
from PyQt5.QtCore import Qt
from minixs.gui.qtcalibrator.calibratorexposureselector import CalibratorExposureSelector
import logging
from logging import StreamHandler
LOGGER_NAME = 'testcalibratorexposureselector'
app = QApplication(sys.argv)
LOG_FORMAT='F- %(asctime)-15s - %(name)s - %(funcName)s- %(levelname)s - %(message)s'

class Test(unittest.TestCase):
    logger = logging.getLogger(LOGGER_NAME)
    handler = StreamHandler()
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(logging.Formatter(LOG_FORMAT))
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    def setUp(self):
        self.old_column_count = 8
        self.filename_olddet = "../data/Cuecal.0001"
        self.new_column_count = 23
        self.filename_newdet = "../data/mini2CrCali.0001"
        self.empty_scan_file = "../data/Cuecal.0001.empty"
        self.not_a_scan_file = "../data/notAScanFile"
        self.ces = CalibratorExposureSelector()

    def tearDown(self):
        pass


    def testAppendRow(self):
        appendButton = self.ces.appendRowButton
        self.assertEqual(self.ces.table.rowCount(), 0, "Test start at zero")
        rowCountBefore = self.ces.table.rowCount()
        # test one Mouse Click
        QTest.mouseClick(appendButton, Qt.LeftButton)
        rowCountAfter = self.ces.table.rowCount()
        self.assertEqual(self.ces.table.rowCount(), 1)
        self.assertEqual(rowCountBefore, rowCountAfter-1, "Test one click append")
        # Test 3 more clicks on append
        rowCountBefore = self.ces.table.rowCount()
        QTest.mouseClick(appendButton, Qt.LeftButton)
        QTest.mouseClick(appendButton, Qt.LeftButton)
        QTest.mouseClick(appendButton, Qt.LeftButton)
        rowCountAfter = self.ces.table.rowCount()
        self.assertEqual(rowCountBefore, rowCountAfter-3, "Test three more clicks append")
        
    def testDeleteRows(self):
        deleteButton = self.ces.deleteRowsButton
        appendButton = self.ces.appendRowButton
        
        #add a blank roww0
        self.assertEqual(self.ces.table.rowCount(), 0, "testDelete add an empty row")
        QTest.mouseClick(appendButton, Qt.LeftButton) 
        
        self.assertEqual(self.ces.table.rowCount(), 1, "testDelete add an empty row")
        
        # now try removing it:
        self.ces.table.setRangeSelected(QTableWidgetSelectionRange(0,0, 0, 1), True)
        selectedRanges = self.ces.table.selectedRanges()
        self.logger.debug ("selectedRanges %s" % selectedRanges)
        self.logger.debug ("selectedRanges.rowCount() %s" % selectedRanges[0].rowCount())
        self.logger.debug ("selectedRanges.topRow() %s" % selectedRanges[0].topRow())
        self.logger.debug ("selectedRanges.leftCoolumn() %s" % selectedRanges[0].leftColumn())
        self.logger.debug ("selectedRanges.bottomRow() %s" % selectedRanges[0].bottomRow())
        self.logger.debug ("selectedRanges.rightColumn() %s" % selectedRanges[0].rightColumn())
        QTest.mouseClick(deleteButton, Qt.LeftButton) 
        
        self.assertEqual(self.ces.table.rowCount(), 0, "testDelete add an empty row rowCount %s, expected %s" % (self.ces.table.rowCount(), 0))
        
if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    logging.basicConfig(stream=sys.stderr)
    logging.getLogger(LOGGER_NAME).setLevel(logging.DEBUG)
    unittest.main()


import numpy as np

SCAN_COLUMN_WIDTH = 21

class ScanFile:
  def __init__(self, filename=None):
    if filename:
      self.load(filename)

  def load(self, filename, headers_only=False):
    self.filename = filename

    #read in headers
    with open(filename) as f:
      self.headers = []
      for line in f:
        if line[0] != '#':
          # rewind to beginning of this line
#           f.seek(-len(line), 1)
          break

        # strip off '#' and newline
        self.headers.append(line)

    column_names = self.headers[-1]
    self.columns = [column_names[i:i+SCAN_COLUMN_WIDTH].strip() for i in range(1,len(column_names)-2, SCAN_COLUMN_WIDTH)]

    # load in data
    if not headers_only:
      self.data = np.loadtxt(filename)

      # reshape single column
      if len(self.data.shape) == 1:
        self.data.shape = (self.data.shape[0], 1)

  def save(self, filename=None, fmt=None):
    if not filename:
      filename = self.filename

    assert filename, "A filename must be provided in order to save"

    if fmt is None:
      fmt = '%%%d.8f' % SCAN_COLUMN_WIDTH

    with open(filename, 'w') as f:
      for line in self.headers:
        f.write(line)

      # numpy cannot save with dos newlines...
      #savetxt(f, self.data, fmt=fmt, delimiter=' ')

      # so, save by hand
      for row in self.data:
        for val in row:
          f.write(fmt % val)
          f.write(' ')
        f.write('\r\n')




import wx
import wx.lib.newevent
from wx.lib.sized_controls import SizedScrolledPanel

from minixs.gui.mouse_event import *
from minixs import METHOD_ENTER_STR
import logging
logger = logging.getLogger(__name__)

EventCoords, EVT_COORDS = wx.lib.newevent.NewCommandEvent()


class ScrolledImageView(SizedScrolledPanel):
  def __init__(self, *args, **kwargs):
    SizedScrolledPanel.__init__(self, args[0], wx.ID_ANY, size=kwargs['size'])

    box = wx.BoxSizer()
    image_view = ImageView(self, *args[1:], **kwargs)
    self.image_view = image_view
    self.SetSizer(box)

    self.SetAutoLayout(True)
    self.SetupScrolling()

class ImageView(SizedScrolledPanel):
  def __init__(self, *args, **kwargs):
    SizedScrolledPanel.__init__(self, *args, **kwargs)
    self.bitmap = None
    self.SetEvtHandlerEnabled(True)
    self.Bind(wx.EVT_PAINT, self.OnPaint)
    self.Bind(wx.EVT_SIZE, self.OnSize)
    self.Bind(wx.EVT_ERASE_BACKGROUND, self.OnEraseBG)
    self.Bind(wx.EVT_LEFT_DOWN, self.OnLeftDown)
    self.Bind(wx.EVT_LEFT_UP, self.OnLeftUp)
    self.Bind(wx.EVT_RIGHT_DOWN, self.OnRightDown)
    self.Bind(wx.EVT_RIGHT_UP, self.OnRightUp)
    self.Bind(wx.EVT_MIDDLE_DOWN, self.OnMiddleDown)
    self.Bind(wx.EVT_MIDDLE_UP, self.OnMiddleUp)
    self.Bind(wx.EVT_MOTION, self.OnMotion)
    self.Bind(wx.EVT_ENTER_WINDOW, self.OnEnterWindow)
    self.Bind(wx.EVT_LEAVE_WINDOW, self.OnLeaveWindow)
    self.Bind(wx.EVT_MOUSEWHEEL, self.OnMouseWheel)
    
    self.tools = []
    self.active_tool = None

    self.panning = False

    self.zoom_on_wheel = False
    self.zoom = 1
    self.zoom_min = -10
    self.zoom_max = 10
    self.set_zoom_delay = 10
    self.set_zoom_timeout = None

    self.pan = [0, 0]
    self.pan_min = [0,0]
    self.pan_max = [0,0]
    self.SetAutoLayout(1)
    self.SetupScrolling()

  def GetBitmapSize(self):
    if self.bitmap:
      return self.raw_image.GetSize()
    else:
      return (0,0)

  def SetPixels(self, pixels, colormap):
    logger.debug(METHOD_ENTER_STR % ((pixels, colormap),))
    if pixels is None:
      logger.debug("Pixels is None")
      self.raw_image = None
      self.bitmap = None
      self.Refresh()
    else:
      logger.debug("Image size %s", pixels.shape)
#       self.SetSize(wx.Size(pixels.shape[1], pixels.shape[0]))
#       self.SetMaxSize(wx.Size(pixels.shape[1], pixels.shape[0]))
      self.Refresh()
      
      h,w = pixels.shape[0:2]
      logger.debug("pixels.shape %s" % (pixels.shape,))
      p = colormap(pixels, bytes=True)[:,:,0:3]
      self.raw_image = wx.ImageFromBuffer(w, h, p.tostring())
      self.SetZoom(self.zoom, force=True, immediate=True)
      self.Refresh()

  def _SetZoomActual(self):
    zoom = self.zoom
    logger.debug("Raw Image Size %d %d" % (self.raw_image.Height, self.raw_image.Width))
    if zoom > 0:
      w = self.raw_image.Width * zoom
      h = self.raw_image.Height * zoom
    elif zoom < 0:
      w = -self.raw_image.Width / zoom
      h = -self.raw_image.Height / zoom
    else:
      raise ValueError("Zoom level cannot be 0.")

    self.zoom = zoom

    scaled = self.raw_image.Scale(w, h)

    self.bitmap = wx.BitmapFromImage(scaled)
    self.UpdatePanBounds()
    self.Refresh()
    self.set_zoom_timeout = None


  def SetZoom(self, zoom, force=False, immediate=False):
    if zoom > self.zoom_max:
      zoom = self.zoom_max

    if zoom < self.zoom_min:
      zoom = self.zoom_min

    if not force and zoom == self.zoom:
      return

    self.zoom = zoom

    if immediate:
      self._SetZoomActual()
    elif not self.set_zoom_timeout:
      self.set_zoom_timeout = wx.CallLater(self.set_zoom_delay, self._SetZoomActual)

  def SetPan(self, pan_x = None, pan_y = None):
    if pan_x is not None:
      self.pan[0] = pan_x

    if pan_y is not None:
      self.pan[1] = pan_y

    self.Refresh()

  def UpdatePanBounds(self):
    if not self.bitmap:
      self.pan_min = [0,0]
      self.pan_max = [0,0]
      return

    px, py = self.pan
    imw, imh = self.bitmap.GetSize()
    bw, bh = self.GetSize()

    if bw < imw:
      xmin = bw - imw
      xmax = 0
    else:
      xmin = 0
      xmax = bw - imw

    if bh < imh:
      ymin = bh - imh
      ymax = 0
    else:
      ymin = 0
      ymax = bh - imh

    self.pan_min = [xmin, ymin]
    self.pan_max = [xmax, ymax]

    if px < xmin: px = xmin
    if px > xmax: px = xmax
    if py < ymin: py = ymin
    if py > ymax: py = ymax

    self.SetPan(px, py)

  def CoordBitmapToScreen(self,x,y):
    px, py = self.pan
    if self.zoom > 0:
      return (x*self.zoom + px, y*self.zoom + py)
    else:
      return (-x/float(self.zoom) + px, -y/float(self.zoom) + py)

  def CoordScreenToBitmap(self,sx,sy):
    px, py = self.pan
    sx -= px
    sy -= py
    if self.zoom > 0:
      return (sx/float(self.zoom), sy/float(self.zoom))
    else:
      return (-sx*self.zoom, -sy*self.zoom)

  def PostEventCoords(self, x, y):
    evt = EventCoords(self.Id, x=x, y=y)
    wx.PostEvent(self, evt)

  def OnLeftDown(self, evt):
    if not self.bitmap:
      return

    if mouse_event_modifier_mask(evt) & MOD_CTRL:
      self.panning = True
      self.pan_prev = None

    for tool in self.tools:
      if tool.active:
        tool.OnLeftDown(evt)

  def OnLeftUp(self, evt):
    self.panning = False

    if not self.bitmap:
      return

    for tool in self.tools:
      if tool.active:
        tool.OnLeftUp(evt)

  def OnRightDown(self, evt):
    if not self.bitmap:
      return

    for tool in self.tools:
      if tool.active:
        tool.OnRightDown(evt)

  def OnRightUp(self, evt):
    if not self.bitmap:
      return

    for tool in self.tools:
      if tool.active:
        tool.OnRightUp(evt)

  def OnMiddleDown(self, evt):
    if not self.bitmap:
      return

    for tool in self.tools:
      if tool.active:
        tool.OnMiddleDown(evt)

  def OnMiddleUp(self, evt):
    if not self.bitmap:
      return

    for tool in self.tools:
      if tool.active:
        tool.OnMiddleUp(evt)

  def OnMotion(self, evt):
    if not self.bitmap:
      return

    x, y = evt.GetPosition()
    if self.panning:
      if self.pan_prev:
        self.SetPan(
          self.pan[0] + x - self.pan_prev[0],
          self.pan[1] + y - self.pan_prev[1]
          )
        self.Refresh()
      self.pan_prev = (x,y)

    x, y = self.CoordScreenToBitmap(x, y)
    self.PostEventCoords(x, y)


    for tool in self.tools:
      if tool.active:
        tool.OnMotion(evt)

  def OnEnterWindow(self, evt):
    logging.debug(METHOD_ENTER_STR)
    for tool in self.tools:
      if tool.active:
        tool.OnEnterWindow(evt)

  def OnLeaveWindow(self, evt):
    self.PostEventCoords(None, None)

    for tool in self.tools:
      if tool.active:
        tool.OnLeaveWindow(evt)

  def OnPaint(self, evt):
    dc = wx.BufferedPaintDC(self)
    self.Draw(evt, dc)

  def OnSize(self, evt):
    logger.debug(METHOD_ENTER_STR % (evt.GetSize(), ))
#     self.SetMaxSize(evt.GetSize())
#     self.SetSize(evt.GetSize())
   
    dc = wx.BufferedPaintDC(self)
    self.Draw(evt, dc)
    self.Layout()
    self.Refresh()

  def OnEraseBG(self, evt):
    pass

  def Draw(self, evt, dc):
    dc.Clear()
    if self.bitmap:
      dc.DrawBitmap(self.bitmap, self.pan[0], self.pan[1])

    for tool in self.tools:
      if tool.visible:
        tool.OnPaint(evt, dc)

  def OnMouseWheel(self, evt):
    if not self.zoom_on_wheel:
      return

    if not self.bitmap:
      return

    rot = evt.GetWheelRotation()

    zoom = self.zoom
    if rot > 0:
      zoom += 1
      if zoom == -1:
        zoom = 1

    elif rot < 0:
      zoom -= 1
      if zoom == 0:
        zoom = -2

    self.SetZoom(zoom)

    # pan so that pixel under mouse stays under mouse
    #x,y = evt.GetPosition()

  def SetPan(self, px, py):
    mx,my = self.pan_min
    if px < mx: px = mx
    if py < my: py = my

    mx,my = self.pan_max
    if px > mx: px = mx
    if py > my: py = my

    self.pan = [px,py]
    self.Refresh()

  def AddTool(self, tool):
    self.tools.append(tool)

  def SetActiveTool(self, tool):
    if tool not in self.tools:
      self.AddTool(tool)

    self.active_tool = tool



MOD_NONE  = 0x0
MOD_SHIFT = 0x1
MOD_CTRL  = 0x2
MOD_ALT   = 0x4
MOD_CMD   = 0x8

def mouse_event_modifier_mask(evt):
  mask = 0
  mask |= evt.ShiftDown() * MOD_SHIFT
  mask |= evt.ControlDown() * MOD_CTRL
  mask |= evt.AltDown() * MOD_ALT
  mask |= evt.CmdDown() * MOD_CMD
  return mask




import numpy as np
from scipy.optimize import leastsq

def gauss(x,x0,sigma):
  return np.exp(-(x-x0)**2 / (2*sigma**2))

def gauss_model(params, x):
  A,x0,sigma = params
  return A*gauss(x,x0,sigma)

def gauss_error(params, *args):
  x,y = args
  return gauss_model(params,x) - y

def gauss_leastsq(data, guess):
  """
  Fit a single gaussian to data

  Parameters
  ----------
    data: tuple containing (x,y) data point vectors
    guess: tuple containing starting (amplitude, mean, stddev) values
  """
  return leastsq(gauss_error, guess, data)



import wx

class MenuFrame(wx.Frame):
  def __init__(self, *args, **kwargs):

    if 'menu_info' in kwargs.keys():
      menu_info = kwargs['menu_info']
      del(kwargs['menu_info'])
    else:
      menu_info = None

    wx.Frame.__init__(self, *args, **kwargs)

    if menu_info:
      self.CreateMenuBar(menu_info)

  def CreateMenuBar(self, menu_info):
    menubar = wx.MenuBar()

    for menu_name, items in menu_info:
      menu = wx.Menu()

      for label, id, description in items:

        if label == '':
          menu.AppendSeparator()
        
        else:
          menu_item = menu.Append(id, label, description)

      menubar.Append(menu, menu_name)

    self.SetMenuBar(menubar)



from minixs.calibrate import Calibration
import wx
import os, imp

from minixs.gui.calibrator.calibrator_controller import CalibratorController
from minixs.gui.calibrator.calibrator_view import CalibratorView
from minixs.gui.calibrator.calibrator_const import *

from minixs import filter
from minixs.gui import filter_view

from glob import glob
import logging
logger = logging.getLogger(__name__)

class CalibratorModel(Calibration):
  def __init__(self):
    Calibration.__init__(self)

class CalibratorApp(wx.App):
  def __init__(self, *args, **kwargs):
    logger.error("Starting CalibratorApp")
    wx.App.__init__(self, *args, **kwargs)

    self.SetAppName('minixs')

    self.setup_configuration()
    self.load_filter_plugins()

    model = CalibratorModel()
    view = CalibratorView(None, ID_MAIN_FRAME, "miniXS Calibrator")
    controller = CalibratorController(view, model)

    view.Show()

  def setup_configuration(self):
    # setup data directory
    datadir = wx.StandardPaths.Get().GetUserDataDir()
    logger.debug ("Datadir: %s" % datadir)
    self.datadir = datadir
    if not os.path.exists(datadir):
      os.makedirs(datadir)

    # create config object and set as global config obj
    conffile = os.path.join(datadir, 'config')
    conf = wx.FileConfig(localFilename=conffile)
    conf.SetRecordDefaults(True)
    wx.Config.Set(conf)

  def load_filter_plugins(self):
    conf = wx.Config.Get()
    filterdir = conf.Read('calibrator/filter_dir', os.path.join(self.datadir, 'filters'))
    if not os.path.exists(filterdir):
      os.makedirs(filterdir)

    # load up all .py files in filterdir
    filter_files = glob(os.path.join(filterdir, '*.py'))
    for filename in filter_files:
      modname = os.path.splitext(os.path.split(filename)[-1])[0]
      m = imp.load_source(modname, filename)
      if hasattr(m, 'register_filters'):
        m.register_filters()
      if hasattr(m, 'register_filter_views'):
        m.register_filter_views()

def main():
  # register built in filters
  # XXX these should be split out into a system filter directory...
  for f in filter.REGISTRY:
    if hasattr(f, 'view_name'):
      view_class = getattr(filter_view, f.view_name)
      filter_view.register(f, view_class)

  # set up application
  app = CalibratorApp()

  # run it
  app.MainLoop()



FILE_UNKNOWN     = -1
FILE_CALIBRATION = 0
FILE_XTALS       = 1
FILE_XES         = 2
FILE_RIXS        = 3

FILE_TYPES = [
    'calibration matrix',
    'crystal boundaries',
    'XES spectrum',
    'RIXS spectrum'
    ]

def determine_filetype(path):
  with open(path) as f:
    line = f.readline()
    if len(line) == 0 or line[0] != "#":
      return FILE_UNKNOWN

    s = line[2:]

    if not s.lower().startswith('minixs'):
      return 0

    t = s[7:].strip()
    if t in FILE_TYPES:
      return FILE_TYPES.index(t)

    return FILE_UNKNOWN

class InvalidFileError(Exception):
  pass



"""
views for various filter types
"""

import wx


EventFilterChanged, EVT_FILTER_CHANGED = wx.lib.newevent.NewCommandEvent()

class FilterView(wx.Panel):
  def __init__(self, *args, **kwargs):
    if 'filter' in kwargs.keys():
      self.filter = kwargs['filter']
      del(kwargs['filter'])

    wx.Panel.__init__(self, *args, **kwargs)

  def SetValue(self, val):
    pass

  def GetValue(self):
    pass

  def Enable(self, enable=True):
    wx.Panel.Enable(self, enable)
    self.ctrl.Enable(enable)

  def post_event_filter_changed(self):
    """
    Send event indicating that filter has changed
    """
    evt = EventFilterChanged(self.Id, filter=self.filter)
    wx.PostEvent(self, evt)

class IntFilterView(FilterView):
  def __init__(self, *args, **kwargs):
    FilterView.__init__(self, *args, **kwargs)

    id = args[1]
    spin = wx.SpinCtrl(self, id, '', max=100000)
    self.ctrl = spin

    self.ctrl.Bind(wx.EVT_SPINCTRL, self.OnSpinBox, id=id)

  def SetValue(self, val):
    self.ctrl.SetValue(int(val))

  def GetValue(self):
    return self.ctrl.GetValue()

  def OnSpinBox(self, evt):
    self.post_event_filter_changed()

class StringFilterView(FilterView):
  def __init__(self, *args, **kwargs):
    FilterView.__init__(self, *args, **kwargs)

    id = args[1]
    text = wx.TextCtrl(self, id, '', style=wx.TE_PROCESS_ENTER)
    self.ctrl = text
    self.lastval = ''

    self.ctrl.Bind(wx.EVT_TEXT_ENTER, self.OnTextBox, id=id)
    self.ctrl.Bind(wx.EVT_KILL_FOCUS, self.OnTextBox, id=id)

  def SetValue(self, val):
    self.ctrl.SetValue(val)

  def GetValue(self):
    return self.ctrl.GetValue()

  def OnTextBox(self, evt):
    newval = self.ctrl.GetValue()
    if self.lastval == newval:
      return

    self.lastval = newval 
    self.post_event_filter_changed()

class ChoiceFilterView(FilterView):
  def __init__(self, *args, **kwargs):
    FilterView.__init__(self, *args, **kwargs)

    id = args[1]
    choice = wx.Choice(self, id, choices=self.filter.CHOICES)
    self.ctrl = choice 

    self.ctrl.Bind(wx.EVT_CHOICE, self.OnChoice, id=id)

  def SetValue(self, val):
    self.ctrl.SetSelection(int(val))

  def GetValue(self):
    return self.ctrl.GetSelection()

  def OnChoice(self, evt):
    self.post_event_filter_changed()

REGISTRY = []

def register(fltr, view):
  global REGISTRY
  REGISTRY.append((fltr, view, wx.NewId()))

def filter_ids():
  global REGISTRY
  return [id for flter, view, id in REGISTRY]



'''
 Copyright (c) 2019, UChicago Argonne, LLC
 See LICENSE file.
'''

from PyQt5.Qt import QPushButton, QLineEdit, QLabel, QHBoxLayout, QVBoxLayout,\
    QDialog, QFileDialog, QWidget, QComboBox, QSpacerItem,\
    QTableWidgetItem, QMessageBox, Qt
from PyQt5.QtCore import pyqtSignal, QModelIndex
from minixs import METHOD_ENTER_STR
from minixs.config.minixsconfigparser import MinixsConfigParser, FILE_SECTION,\
    LAST_DIRECTORY_STR
from minixs.gui.qtcalibrator.model.exposurelistmodel import ExposureListModel,\
    EXPOSURE_COLUMN_NUMBER, ENERGY_COLUMN_NUMBER
from minixs.gui.qtcalibrator.view.exposurelistview import ExposureListView
from os.path import realpath
import logging
import minixs.gui.util
import minixs.misc
import os
from PyQt5.QtWidgets import QDialogButtonBox
logger = logging.getLogger(__name__)

SELECT_ENERGY_TITLE = "Select Energy File"
EMPTY_STR = ""
# Strings used as GUI labels
CANCEL_LABEL = "Cancel"
COLUMN_LABEL = "Column:"
DATASET_NAME_LABEL = "DatasetName: "
DONE_LABEL = "Done"
READ_ENERGIES_LABEL = "Read Energies..."
SELECT_EXPOSURES_LABEL = "Select Exposures..."
APPEND_ROW_LABEL = "AppendRow"
DELETE_ROWS_LABEL = "Delete Row(s)"
CLEAR_ENERGIES_LABEL = "Clear Energies"
CLEAR_EXPOSURES_LABEL = "Clear Exposures"
READ_ENERGY_DIALOG_TITLE = "Open Energy File"
TEXT_FILE_LABEL = "Text File"
# Strings used to name object signals
ENERGIES_LIST_CHANGED_SIGNAL = "exposureListChangedSignal"
EXPOSURE_LIST_CHANGED_SIGNAL = "energiesListChangedSignal"
TABLE_APPEND_ROW_SIGNAL = "tableAppendRow"
TABLE_DELETE_ROWS_SIGNAL = "tableRemoveRows"
CLEAR_ENERGIES_SIGNAL = "clearEnergies"
CLEAR_EXPOSURES_SIGNAL = "clearExposures"

class SelectEnergiesDialog(QDialog):
    '''
    Dialog for selecting file containing energies and selecting which column contains the energies
    '''

    def __init__(self, *args, **kwargs):
        super(SelectEnergiesDialog, self).__init__(*args, **kwargs)
        self.setWindowTitle(SELECT_ENERGY_TITLE)
        self.config = MinixsConfigParser()
        self.filename = EMPTY_STR
        vLayout = QVBoxLayout()
        hLayout = QHBoxLayout()
        #ROW 1
        label = QLabel(TEXT_FILE_LABEL)
        self.fileNameTxt = QLineEdit()
        fileButton = QPushButton("...")
        fileButton.clicked.connect(self.onSelectFile)
        self.fileNameTxt.editingFinished.connect(self.onFilenameEditFinished)
        hLayout.addWidget(label)
        hLayout.addWidget(self.fileNameTxt)
        hLayout.addWidget(fileButton)
        vLayout.addLayout(hLayout)
        #ROW 2
        hLayout = QHBoxLayout()
        label = QLabel(COLUMN_LABEL)
        self.columnSelect = QComboBox()
        hLayout.addWidget(label)
        hLayout.addWidget(self.columnSelect)
        vLayout.addLayout(hLayout)
        #ROW 3
        hLayout = QHBoxLayout()
        spacer = QSpacerItem(100, 10)
        hLayout.addItem(spacer)
        buttons = QDialogButtonBox(QDialogButtonBox.Ok |
                                   QDialogButtonBox.Cancel, 
                                   Qt.Horizontal,self)
        
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        vLayout.addWidget(buttons)
        self.setLayout(vLayout)
    
    def cancelSelectEnergies(self):
        logger.debug(METHOD_ENTER_STR)   
        self.finished.emit(self.rejected)

    def doneSelectEnergies(self):
        logger.debug(METHOD_ENTER_STR)
        self.finished.emit(self.accepted)
        
    def onSelectFile(self):
        lastPath = self.config.getLastFilePath()
        filename, _ = QFileDialog.getOpenFileName(None, 
                                        caption=READ_ENERGY_DIALOG_TITLE, 
                                        directory=lastPath)
        logger.debug("Filename %s" % filename)
        self.processNewFilename(filename)

    def onFilenameEditFinished(self,):
        filename = self.fileNameTxt.text()
        file_exists = os.path.isfile(filename)
        if file_exists:
            self.processNewFilename(filename)
        else:
            if filename != "":
                QMessageBox.warning(self, "File %s does not exist" % filename)
             
    def processNewFilename(self, filename):
        if filename != "":
            columns = None
            try:
                columns = minixs.gui.util.read_scan_column_names(filename)
            except IndexError:
                QMessageBox.warning(self, "Error", \
                                    (("Trouble reading %s.  Make sure that " + \
                                    "this is a proper scanFile.") % filename))
                return
            
            if columns:
                self.columnSelect.clear()
                self.columnSelect.addItems(columns)
    
            for col in columns:
                if 'energy' in col.lower():
                    self.columnSelect.setCurrentText(col)
            self.filename = filename
            self.fileNameTxt.setText(self.filename)
            self.config.set(FILE_SECTION, LAST_DIRECTORY_STR, realpath(self.filename))
            self.config.updateFile()
        else:
            pass

        
    def getSelectedColumn(self):
        if self.filename != EMPTY_STR:
            columnData = minixs.misc.read_scan_info(self.filename, \
                            [self.columnSelect.currentIndex()])[0]
            return columnData
        else:
            return []
                
class CalibratorExposureSelector(QWidget):
    '''
    class to set up a table to list energies and image files for doing 
    the energy calibration.
    '''

    exposureListChanged = pyqtSignal(name=EXPOSURE_LIST_CHANGED_SIGNAL)    
    energiesListChanged = pyqtSignal(name=ENERGIES_LIST_CHANGED_SIGNAL)
    tableAppendRow = pyqtSignal(name=TABLE_APPEND_ROW_SIGNAL)
    tableDeleteRows = pyqtSignal(list, name=TABLE_DELETE_ROWS_SIGNAL)
    clearExposures = pyqtSignal(name=CLEAR_EXPOSURES_SIGNAL)
    clearEnergies = pyqtSignal(name=CLEAR_ENERGIES_SIGNAL)
    
    def __init__(self, parent=None):
        '''
        Constructor
        '''
        super(CalibratorExposureSelector, self).__init__(parent)
        self.config = MinixsConfigParser()
        
        vLayout = QVBoxLayout()
        hLayout = QHBoxLayout()
        label = QLabel(DATASET_NAME_LABEL)
        self.datasetNameTxt = QLineEdit()
        hLayout.addWidget(label)
        hLayout.addWidget(self.datasetNameTxt)
        vLayout.addLayout(hLayout)
        self.model = None
        
        hLayout = QHBoxLayout()
        
        self.tableView = ExposureListView()
        hLayout.addWidget(self.tableView)
        
        buttonLayout = QVBoxLayout()
        
        self.readEnergiesButton = QPushButton(READ_ENERGIES_LABEL)
        self.selectExposuresButton = QPushButton(SELECT_EXPOSURES_LABEL)
        self.appendRowButton = QPushButton(APPEND_ROW_LABEL)
        self.deleteRowsButton = QPushButton(DELETE_ROWS_LABEL)
        self.clearEnergiesButton = QPushButton(CLEAR_ENERGIES_LABEL)
        self.clearExposuresButton = QPushButton(CLEAR_EXPOSURES_LABEL)
        self.selectExposuresButton.setEnabled(False)
        self.deleteRowsButton.setEnabled(False)
        self.clearEnergiesButton.setEnabled(False)
        self.clearExposuresButton.setEnabled(False)
        buttonLayout.addWidget(self.readEnergiesButton)
        buttonLayout.addWidget(self.selectExposuresButton)
        buttonLayout.addWidget(self.appendRowButton)
        buttonLayout.addWidget(self.deleteRowsButton)
        buttonLayout.addWidget(self.clearEnergiesButton)
        buttonLayout.addWidget(self.clearExposuresButton)
        
        hLayout.addLayout(buttonLayout)
        vLayout.addLayout(hLayout)
        self.setLayout(vLayout)
        
        self.readEnergiesButton.clicked.connect(self.onReadEnergies)
        self.selectExposuresButton.clicked.connect(self.onSelectExposures)
        self.appendRowButton.clicked.connect(self.onAppendRow)
        self.deleteRowsButton.clicked.connect(self.onDeleteRows)
        self.clearEnergiesButton.clicked.connect(self.onClearEnergies)
        self.clearExposuresButton.clicked.connect(self.onClearExposures)

    def appendRows(self, energies, filenames):
#         self.tableView.model().setRowCount(len(energies))
        row = 0
        for energy, filename in zip(energies, filenames):
            self.tableView.model().appendEnergy(energy)
            self.tableView.model().appendExposure(filename)
            row += 1
        self.exposureListChanged.emit()
       
#     def getExposureList(self):
#         exposures = []
#         for row in range(self.tableView.model().rowCount()):
#             expName = str(self.tableView.model().item(row, 1).text())
#             if expName == EMPTY_STR:
#                 logger.debug("item contains empty string")
#             exposures.append(expName)
#         return exposures
#             
    def onReadEnergies(self):
        '''
        Method for signal to read Energies
        '''
        logger.debug(METHOD_ENTER_STR)
        energySelectDialog = SelectEnergiesDialog()
        status = energySelectDialog.exec_()
#         status = energySelectDialog.result()
        logger.debug ("Energy Select Dialog Status %s" % status)
        #logger.debug("energySelector Accepted: %s" % energySelectDialog.accepted())
        if status == QDialog.Rejected:
            return
        selectedEnergies =  energySelectDialog.getSelectedColumn()
        logger.debug("selectedEnergies %s" % (selectedEnergies,))
        if len(selectedEnergies) == 0:            # No Energies were loaded
            pass
        else:
            for energy in selectedEnergies:
                self.tableView.model().appendEnergy(energy)
            
        if (self.tableView.model().rowCount() > 0):
            self.selectExposuresButton.setEnabled(True)
            self.deleteRowsButton.setEnabled(True)
            self.clearEnergiesButton.setEnabled(True)
            self.clearExposuresButton.setEnabled(True)
            self.energiesListChanged.emit()
        else:
            self.selectExposuresButton.setEnabled(False)
            self.deleteRowsButton.setEnabled(False)
            self.clearEnergiesButton.setEnabled(False)
            self.clearExposuresButton.setEnabled(False)
    
    def onSelectExposures(self):
        '''
        Method for signal to select exposures
        '''
        logger.debug(METHOD_ENTER_STR)
        lastPath = self.config.getLastFilePath()
        filenames, _ = QFileDialog.getOpenFileNames(None, 
                                         caption=READ_ENERGY_DIALOG_TITLE, 
                                         directory=lastPath)
        logger.debug("Filename %s" % (filenames,))
        if len(filenames) == 0:
            return
        for exposureFile in filenames:
            self.tableView.model().appendExposure(exposureFile)
            
        self.exposureListChanged.emit()
        
    def onAppendRow(self):
        '''
        Method to append a row
        '''
        logger.debug(METHOD_ENTER_STR)
        rowCount = self.tableView.model().rowCount()
#         self.tableView.model().setRowCount(rowCount+1)
        self.tableView.model().insertRow(rowCount)
        if self.tableView.model().rowCount() > 0:
            self.deleteRowsButton.setEnabled(True)
        else:
            self.deleteRowsButton.setEnabled(False)
        self.tableAppendRow.emit()
        
    def onDeleteRows(self):
        '''
        Method to delete rows
        '''
        logger.error(METHOD_ENTER_STR)
        # Append 
        selectedIndexes = self.tableView.selectedIndexes()
        logger.debug("selectedRanges %s" % selectedIndexes)
        selectedRows = set()
        for index in selectedIndexes:
            selectedRows.add(index.row())
        if len(selectedRows) > 0:
            selectedRows = list(selectedRows)
            selectedRows.sort(reverse=True)
        else:
            return
        for row in selectedRows:
            logger.debug("Removing Row %s" % row)
            self.tableView.model().removeRow(row)
        logger.debug("# of rows %s" % self.tableView.model().rowCount())
        if self.tableView.model().rowCount() > 0:
            self.deleteRowsButton.setEnabled(True)
        else:
            self.deleteRowsButton.setEnabled(False)
        self.tableDeleteRows.emit(selectedRows)
        
        
    def onClearEnergies(self):
        '''
        method to clear energies
        '''
        logger.debug(METHOD_ENTER_STR)
        for row in range(self.tableView.model().rowCount()):
            index = self.tableView.model().index(row, ENERGY_COLUMN_NUMBER, QModelIndex())
            self.tableView.model().setData(index, None)
        self.clearEnergies.emit()
        
    def onClearExposures(self):
        '''
        method to clear exposure list
        '''
        logger.debug(METHOD_ENTER_STR)
        for row in range(self.tableView.model().rowCount()):
            index = self.tableView.model().index(row, EXPOSURE_COLUMN_NUMBER,QModelIndex())
            self.tableView.model().setData(index, None)
        
        self.clearExposures.emit()
        
    def getData(self):
        '''
        retrieve data from the table for energies and filenames, if there is 
        a matching filename for each energy then return True for valid,
        otherwise return false for Valid
        
        '''
        energies = []
        files = []
        valid = True
        fileName = ""
        energy = ""
        for row in range(self.tableView.model().rowCount()):
            index = self.tableView.model().index(row, ENERGY_COLUMN_NUMBER, QModelIndex())
            energyItem = self.tableView.model().data(index)
            logger.debug("energyItem %s" % energyItem)
            logger.debug(type("energyItem) %s" % type(energyItem) ))
            logger.debug("energyItem is None '%s'" % (energyItem is None))
            logger.debug("len(energyItem) %s" % (len(energyItem)))
            logger.debug("dir(energyItem %s" % dir(energyItem))
            logger.debug("energyItem is None %s" % (energyItem == ''))
            if energyItem is None or energyItem =='None':
                energy = None
            else:
                energy = float(energyItem)
            index = self.tableView.model().index(row, EXPOSURE_COLUMN_NUMBER, QModelIndex())
            fileItem = self.tableView.model().data(index)
            logger.debug("fileItem %s"  % fileItem)
            if fileItem is None or str(fileItem)=='':
                fileName = None
            else:
                fileName = str(fileItem)
            if energy == None and fileName == None:
                continue
            if energy == None or fileName==None:
                valid = False
                continue
            energies.append(float(energy))
            files.append(fileName)
        return (valid, energies, files)
    
    def setButtonsForNoExposures(self):
        '''
        Set buttons on right side for the case that no exposures have been selected.
        '''
        self.selectExposuresButton.setEnabled(True)
        self.deleteRowsButton.setEnabled(True)
        self.clearEnergiesButton.setEnabled(True)
        self.clearExposuresButton.setEnabled(True)
        self.energiesListChanged.emit()
    
    def setButtonsForExposures(self):
        '''
        Set buttons on the right side for the case that images have been added.
        '''
        self.selectExposuresButton.setEnabled(True)
        self.deleteRowsButton.setEnabled(True)
        self.clearEnergiesButton.setEnabled(True)
        self.clearExposuresButton.setEnabled(True)
        self.energiesListChanged.emit()
        
    def setCalibrationModel(self, calibrationModel):
        self.calibrationModel = calibrationModel
        self.exposureListModel = ExposureListModel()
        self.exposureListModel.setCalibration(self.calibrationModel)
        self.tableView.setModel(self.exposureListModel)
        